import shutil
import tempfile

import pytest
import pyarrow as pa
import pandas as pd
import polars as pl
import numpy as np
import ray.data as rd
import daft

import deltacat.catalog.main.impl as catalog
from deltacat.catalog import get_catalog_properties
from deltacat.storage.model.schema import (
    Schema,
    Field,
)
from deltacat.storage.model.types import SchemaConsistencyType
from deltacat.storage.model.sort_key import SortKey, SortScheme, SortOrder, NullOrder
from deltacat.storage.model.types import LifecycleState
from deltacat.exceptions import (
    TableAlreadyExistsError,
    TableNotFoundError,
    TableValidationError,
    SchemaValidationError,
)
from deltacat.types.tables import TableWriteMode, TableProperty, SchemaEvolutionMode
from deltacat.types.media import ContentType


@pytest.fixture(scope="class")
def catalog_setup():
    """Setup and teardown for the catalog test environment."""
    temp_dir = tempfile.mkdtemp()
    catalog_properties = get_catalog_properties(root=temp_dir)
    yield temp_dir, catalog_properties

    # Teardown
    shutil.rmtree(temp_dir)


@pytest.fixture(scope="function")
def test_namespace(catalog_setup):
    """Create a test namespace for each test."""
    _, catalog_properties = catalog_setup
    namespace_name = "test_table_namespace"

    if not catalog.namespace_exists(namespace_name, inner=catalog_properties):
        catalog.create_namespace(
            namespace=namespace_name,
            properties={"description": "Test Table Namespace"},
            inner=catalog_properties,
        )

    return namespace_name, catalog_properties


@pytest.fixture
def sample_arrow_schema():
    """Create a sample PyArrow schema for testing."""
    return pa.schema(
        [
            pa.field("id", pa.int64()),
            pa.field("name", pa.string()),
            pa.field("value", pa.float64()),
        ]
    )


@pytest.fixture
def sample_sort_keys():
    """Create a sample sort scheme for testing."""
    return SortScheme(
        keys=[
            SortKey.of(
                key=["id"], sort_order=SortOrder.ASCENDING, null_order=NullOrder.AT_END
            ),
        ]
    )


class TestCatalogTableOperations:
    """Test catalog table operations including table creation, existence checks, etc."""

    @classmethod
    def setup_class(cls):
        cls.temp_dir = tempfile.mkdtemp()
        cls.catalog_properties = get_catalog_properties(root=cls.temp_dir)

        # Create a test namespace
        cls.test_namespace = "test_write_operations"
        catalog.create_namespace(
            namespace=cls.test_namespace,
            inner=cls.catalog_properties,
        )

    @classmethod
    def teardown_class(cls):
        shutil.rmtree(cls.temp_dir)

    def test_create_table(self, test_namespace, sample_arrow_schema, sample_sort_keys):
        """Test creating a table with schema and properties"""
        namespace_name, catalog_properties = test_namespace
        table_name = "test_create_table"

        # Create a schema
        schema = Schema(arrow=sample_arrow_schema)

        # Create table properties
        table_properties = {"owner": "test-user", "department": "engineering"}

        # Create namespace properties
        namespace_properties = {"description": "Test Namespace"}

        # Create the table
        table_definition = catalog.create_table(
            table=table_name,
            namespace=namespace_name,
            schema=schema,
            sort_keys=sample_sort_keys,
            table_description="Test table for unit tests",
            table_properties=table_properties,
            namespace_properties=namespace_properties,
            inner=catalog_properties,
        )

        # Verify table was created
        assert catalog.table_exists(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        table = table_definition.table
        table_version = table_definition.table_version

        # Verify table definition properties
        assert table_version.table_name == table_name
        assert table_version.namespace == namespace_name
        assert table_version.description == "Test table for unit tests"
        assert table_version.state == LifecycleState.ACTIVE
        assert table.properties.get("owner") == "test-user"
        assert table.properties.get("department") == "engineering"
        assert table_version.schema.arrow.names == sample_arrow_schema.names
        assert len(table_version.sort_scheme.keys) == 1
        sort_key_paths = [key[0][0] for key in table_version.sort_scheme.keys]
        assert "id" in sort_key_paths

    def test_create_table_already_exists(self, test_namespace):
        namespace_name, catalog_properties = test_namespace
        table_name = "test_table_exists"

        # Create the table
        catalog.create_table(
            table=table_name,
            namespace=namespace_name,
            table_description="First creation",
            inner=catalog_properties,
        )

        # Verify table exists
        assert catalog.table_exists(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Try to create the same table again, should raise TableAlreadyExistsError
        with pytest.raises(
            TableAlreadyExistsError,
            match=f"Table {namespace_name}.{table_name} already exists",
        ):
            catalog.create_table(
                table=table_name,
                namespace=namespace_name,
                table_description="Second creation attempt",
                inner=catalog_properties,
            )

    def test_create_table_already_exists_no_fail(self, test_namespace):
        """Test creating a table that already exists with fail_if_exists=False"""
        namespace_name, catalog_properties = test_namespace
        table_name = "test_table_exists_no_fail"

        # Create the table with original description
        catalog.create_table(
            table=table_name,
            namespace=namespace_name,
            table_description="Original description",
            inner=catalog_properties,
        )

        assert catalog.table_exists(
            table_name,
            namespace=namespace_name,
            catalog=catalog_properties,
        )

        # Create the same table with fail_if_exists=False
        table_definition = catalog.create_table(
            table=table_name,
            namespace=namespace_name,
            table_description="Updated description",
            fail_if_exists=False,
            inner=catalog_properties,
        )

        table = table_definition.table

        assert table.table_name == table_name
        assert table.namespace == namespace_name
        # Ensure description is unchanged
        assert table.description == "Original description"

    def test_drop_table(self, test_namespace):
        namespace_name, catalog_properties = test_namespace
        table_name = "test_drop_table"

        # Create the table
        catalog.create_table(
            table=table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Verify table exists
        assert catalog.table_exists(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Drop the table
        catalog.drop_table(
            table=table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Verify table no longer exists
        assert not catalog.table_exists(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

    def test_drop_table_not_exists(self, test_namespace):
        namespace_name, catalog_properties = test_namespace
        table_name = "nonexistent_table"

        # Verify table doesn't exist
        assert not catalog.table_exists(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Try to drop the table, should raise TableNotFoundError
        with pytest.raises(TableNotFoundError, match=table_name):
            catalog.drop_table(
                table=table_name,
                namespace=namespace_name,
                inner=catalog_properties,
            )

    def test_rename_namespace(self, test_namespace):
        namespace_name, catalog_properties = test_namespace
        original_name = "test_original_table"
        new_name = "test_renamed_namespace"

        # Create the table with original name
        catalog.create_table(
            table=original_name,
            namespace=namespace_name,
            table_description="Table to in namespace to be renamed",
            inner=catalog_properties,
        )

        # Verify original table exists
        assert catalog.table_exists(
            original_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Rename the namespace
        catalog.alter_namespace(
            namespace=namespace_name,
            new_namespace=new_name,
            inner=catalog_properties,
        )

        # Verify new namespace exists and old namespace doesn't
        assert catalog.namespace_exists(new_name, inner=catalog_properties)
        assert not catalog.namespace_exists(namespace_name, inner=catalog_properties)

        # Verify we can still discover the table in the new namespace
        assert catalog.table_exists(
            original_name,
            namespace=new_name,
            inner=catalog_properties,
        )

    def test_rename_table(self, test_namespace):
        namespace_name, catalog_properties = test_namespace
        original_name = "test_original_table"
        new_name = "test_renamed_table"

        # Create the table with original name
        catalog.create_table(
            table=original_name,
            namespace=namespace_name,
            table_description="Table to be renamed",
            inner=catalog_properties,
        )

        # Verify original table exists
        assert catalog.table_exists(
            original_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Rename the table
        catalog.rename_table(
            table=original_name,
            new_name=new_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Verify new table exists and old table doesn't
        assert catalog.table_exists(
            new_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )
        assert not catalog.table_exists(
            original_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

    def test_rename_table_not_exists(self, test_namespace):
        namespace_name, catalog_properties = test_namespace
        original_name = "nonexistent_table"
        new_name = "test_renamed_nonexistent"

        # Verify table doesn't exist
        assert not catalog.table_exists(
            original_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Try to rename the table, should raise TableNotFoundError
        with pytest.raises(TableNotFoundError, match=original_name):
            catalog.rename_table(
                table=original_name,
                new_name=new_name,
                namespace=namespace_name,
                inner=catalog_properties,
            )

    def test_table_exists(self, test_namespace):
        namespace_name, catalog_properties = test_namespace
        existing_table = "test_table_exists_check"
        non_existing_table = "nonexistent_table"

        # Create a table
        catalog.create_table(
            table=existing_table,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Check existing table
        assert catalog.table_exists(
            existing_table,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Check non-existing table
        assert not catalog.table_exists(
            non_existing_table,
            namespace=namespace_name,
            inner=catalog_properties,
        )

    def test_create_table_with_default_namespace(self, catalog_setup):
        _, catalog_properties = catalog_setup
        table_name = "test_default_namespace_table"

        # Create table with default namespace
        table_definition = catalog.create_table(
            table=table_name, inner=catalog_properties
        )

        table = table_definition.table
        # Verify table was created in default namespace
        default_ns = catalog.default_namespace()
        assert table.namespace == default_ns
        assert catalog.table_exists(
            table_name,
            namespace=default_ns,
            inner=catalog_properties,
        )

    def test_create_table_with_missing_namespace(self, catalog_setup):
        _, catalog_properties = catalog_setup
        table_name = "test_namespace_not_found_table"
        new_namespace = "nonexistent_namespace"

        # Verify namespace doesn't exist yet
        assert not catalog.namespace_exists(new_namespace, inner=catalog_properties)

        # Try to create table with non-existent namespace
        catalog.create_table(
            table=table_name,
            namespace=new_namespace,
            inner=catalog_properties,
        )

        assert catalog.table_exists(
            table_name,
            namespace=new_namespace,
            inner=catalog_properties,
        )
        assert catalog.namespace_exists(new_namespace, inner=catalog_properties)

    def test_alter_table(self, test_namespace, sample_arrow_schema, sample_sort_keys):
        namespace_name, catalog_properties = test_namespace
        table_name = "test_alter_table"

        # Create initial schema and properties
        schema = Schema.of(schema=sample_arrow_schema)
        initial_properties = {"owner": "original-user", "department": "engineering"}

        # Create the table with initial properties
        table = catalog.create_table(
            table=table_name,
            namespace=namespace_name,
            schema=schema,
            sort_keys=sample_sort_keys,
            table_description="Initial description",
            table_properties=initial_properties,
            inner=catalog_properties,
        )
        old_schema = table.table_version.schema

        # Verify table was created with initial properties
        assert catalog.table_exists(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Create schema update operations to add a new field
        new_field = Field.of(pa.field("count", pa.float64(), nullable=True))
        schema_update = old_schema.update().add_field(new_field)

        # Create updated properties
        updated_properties = {
            "owner": "new-user",
            "department": "data-science",
            "priority": "high",
        }

        # Alter the table with new properties and schema updates
        catalog.alter_table(
            table=table_name,
            namespace=namespace_name,
            schema_updates=schema_update,
            table_description="Updated description",
            table_properties=updated_properties,
            inner=catalog_properties,
        )

        # Get the updated table definition
        updated_table_def = catalog.get_table(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        updated_table = updated_table_def.table
        updated_table_version = updated_table_def.table_version

        # Verify table properties were updated
        assert updated_table_version.description == "Updated description"
        assert updated_table_version.state == LifecycleState.ACTIVE
        assert updated_table.properties.get("owner") == "new-user"
        assert updated_table.properties.get("department") == "data-science"
        assert updated_table.properties.get("priority") == "high"

        # Verify schema was updated with new field
        updated_schema = updated_table_version.schema
        assert updated_schema.field("count") is not None
        assert updated_schema.field("count").arrow.type == pa.float64()
        assert updated_schema.field("count").arrow.nullable is True
        assert (
            updated_schema.field("count").id == 3
        )  # Next sequential ID after id(0), name(1), value(2)

        # Verify schema ID was incremented (proving SchemaUpdate was used)
        assert updated_schema.id == old_schema.id + 1

    def test_alter_table_not_exists(self, test_namespace):
        """Test altering a table that doesn't exist"""
        namespace_name, catalog_properties = test_namespace
        nonexistent_table = "nonexistent_alter_table"

        # Verify table doesn't exist
        assert not catalog.table_exists(
            nonexistent_table,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Try to alter the nonexistent table, should raise TableNotFoundError
        with pytest.raises(TableNotFoundError, match=nonexistent_table):
            catalog.alter_table(
                table=nonexistent_table,
                namespace=namespace_name,
                table_description="Updated description",
                inner=catalog_properties,
            )

    def test_alter_table_with_multiple_schema_operations(
        self, test_namespace, sample_arrow_schema
    ):
        """Test altering a table with multiple schema update operations."""
        namespace_name, catalog_properties = test_namespace
        table_name = "test_alter_table_multiple_ops"

        # Create initial schema
        schema = Schema.of(schema=sample_arrow_schema)
        print("schema.max_field_id", schema.max_field_id)

        # Create the table
        table = catalog.create_table(
            table=table_name,
            namespace=namespace_name,
            schema=schema,
            table_description="Initial description",
            inner=catalog_properties,
        )

        original_schema = table.table_version.schema

        # Create multiple schema update operations
        new_field1 = Field.of(pa.field("count", pa.int64(), nullable=True))
        new_field2 = Field.of(
            pa.field("status", pa.string(), nullable=False),
            past_default="active",
        )

        schema_update = (
            original_schema.update().add_field(new_field1).add_field(new_field2)
        )
        print("original_schema.max_field_id", original_schema.max_field_id)
        print(
            "schema_update.base_schema.max_field_id",
            schema_update.base_schema.max_field_id,
        )

        # Alter the table
        catalog.alter_table(
            table=table_name,
            namespace=namespace_name,
            schema_updates=schema_update,
            table_description="Updated with multiple fields",
            inner=catalog_properties,
        )

        # Get the updated table
        updated_table_def = catalog.get_table(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        updated_schema = updated_table_def.table_version.schema

        # Verify both fields were added
        assert updated_schema.field("count") is not None
        assert updated_schema.field("count").arrow.type == pa.int64()
        assert (
            updated_schema.field("count").id == 3
        )  # Next sequential ID after id(0), name(1), value(2)

        assert updated_schema.field("status") is not None
        assert updated_schema.field("status").arrow.type == pa.string()
        assert (
            updated_schema.field("status").id == 4
        )  # Next sequential ID after count(3)
        assert updated_schema.field("status").past_default == "active"

        # Verify schema ID was incremented
        assert updated_schema.id == original_schema.id + 1

    def test_alter_table_with_remove_operation(self, test_namespace):
        """Test altering a table with field removal (requires allow_incompatible_changes)."""
        namespace_name, catalog_properties = test_namespace
        table_name = "test_alter_table_remove"

        # Create schema with multiple fields
        initial_fields = [
            Field.of(
                pa.field("id", pa.int64(), nullable=False),
                is_merge_key=True,
                field_id=1,
            ),
            Field.of(pa.field("name", pa.string(), nullable=True), field_id=2),
            Field.of(pa.field("temp_field", pa.float64(), nullable=True), field_id=3),
        ]
        schema = Schema.of(initial_fields)

        # Create the table
        table = catalog.create_table(
            table=table_name,
            namespace=namespace_name,
            schema=schema,
            inner=catalog_properties,
        )
        original_schema = table.table_version.schema
        temp_field = original_schema.field("temp_field")
        assert temp_field is not None

        schema_update = original_schema.update(True).remove_field("temp_field")

        catalog.alter_table(
            table=table_name,
            namespace=namespace_name,
            schema_updates=schema_update,
            inner=catalog_properties,
        )

        # If successful, verify the field was removed
        updated_table_def = catalog.get_table(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )
        updated_schema = updated_table_def.table_version.schema

        # temp_field should be removed
        with pytest.raises(KeyError):
            updated_schema.field("temp_field")

        # all other fields should be present
        assert updated_schema.field("id") is not None
        assert updated_schema.field("id").arrow.type == pa.int64()
        assert updated_schema.field("id").id == 1
        assert updated_schema.field("name") is not None
        assert updated_schema.field("name").arrow.type == pa.string()
        assert updated_schema.field("name").id == 2

    def test_alter_table_with_update_operation(self, test_namespace):
        """Test altering a table with field update operation."""
        namespace_name, catalog_properties = test_namespace
        table_name = "test_alter_table_update"

        # Create schema with a field to update
        initial_fields = [
            Field.of(
                pa.field("id", pa.int64(), nullable=False),
                is_merge_key=True,
                field_id=1,
            ),
            Field.of(pa.field("value", pa.int32(), nullable=True), field_id=2),
        ]
        schema = Schema.of(initial_fields)

        # Create the table
        table = catalog.create_table(
            table=table_name,
            namespace=namespace_name,
            schema=schema,
            inner=catalog_properties,
        )

        original_schema = table.table_version.schema

        # Update the value field to int64 (compatible type widening)
        schema_update = original_schema.update().update_field_type("value", pa.int64())

        # Alter the table
        catalog.alter_table(
            table=table_name,
            namespace=namespace_name,
            schema_updates=schema_update,
            inner=catalog_properties,
        )

        # Get the updated table
        updated_table_def = catalog.get_table(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        updated_schema = updated_table_def.table_version.schema

        # Verify field was updated
        assert updated_schema.field("value").arrow.type == pa.int64()
        assert updated_schema.field("value").id == 2

        # Verify schema ID was incremented
        assert updated_schema.id == original_schema.id + 1

    def test_alter_table_with_schema_evolution_disabled(self, test_namespace):
        """Test that alter_table raises TableValidationError when schema evolution is disabled."""
        namespace_name, catalog_properties = test_namespace
        table_name = "test_alter_table_schema_evolution_disabled"

        # Create initial schema
        initial_fields = [
            Field.of(
                pa.field("id", pa.int64(), nullable=False),
                is_merge_key=True,
                field_id=1,
            ),
            Field.of(pa.field("value", pa.int32(), nullable=True), field_id=2),
        ]
        schema = Schema.of(initial_fields)

        # Create table with SCHEMA_EVOLUTION_MODE.DISABLED
        table_properties = {
            TableProperty.SCHEMA_EVOLUTION_MODE: SchemaEvolutionMode.DISABLED
        }

        table = catalog.create_table(
            table=table_name,
            namespace=namespace_name,
            schema=schema,
            table_properties=table_properties,
            inner=catalog_properties,
        )

        original_schema = table.table_version.schema

        # Try to add a new field - this should be blocked
        new_field = Field.of(pa.field("description", pa.string(), nullable=True))
        schema_update = original_schema.update().add_field(new_field)

        # Alter table with schema updates should raise TableValidationError
        with pytest.raises(
            TableValidationError,
            match="Schema evolution is disabled for this table. Please enable schema evolution or remove schema updates.",
        ):
            catalog.alter_table(
                table=table_name,
                namespace=namespace_name,
                schema_updates=schema_update,
                inner=catalog_properties,
            )

        # Verify the schema wasn't changed
        unchanged_table_def = catalog.get_table(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )
        unchanged_schema = unchanged_table_def.table_version.schema

        # Schema should be unchanged
        assert unchanged_schema.id == original_schema.id
        assert len(unchanged_schema.fields) == len(original_schema.fields)

        # Verify the new field was not added
        field_names = [field.arrow.name for field in unchanged_schema.fields]
        assert "description" not in field_names

        # Test that alter_table works without schema_updates even when schema evolution is disabled
        catalog.alter_table(
            table=table_name,
            namespace=namespace_name,
            table_description="Updated description without schema changes",
            inner=catalog_properties,
        )

        # Verify that table description was updated but schema remains unchanged
        final_table_def = catalog.get_table(
            table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )
        assert (
            final_table_def.table_version.description
            == "Updated description without schema changes"
        )
        assert final_table_def.table_version.schema.id == original_schema.id

    def test_drop_with_purge_validation(self, test_namespace):
        """Test that using purge flag raises ValidationError"""
        namespace_name, catalog_properties = test_namespace
        table_name = "test_drop_with_purge"

        # Create the table
        catalog.create_table(
            table=table_name,
            namespace=namespace_name,
            inner=catalog_properties,
        )

        # Try to drop with purge=True, should raise ValidationError
        with pytest.raises(
            NotImplementedError, match="Purge flag is not currently supported"
        ):
            catalog.drop_table(
                table=table_name,
                namespace=namespace_name,
                purge=True,
                inner=catalog_properties,
            )

    def test_create_table_basic(self):
        """Test basic table creation"""
        table_name = "test_create_table_basic"
        schema = Schema.of(
            schema=pa.schema(
                [
                    ("id", pa.int64()),
                    ("name", pa.string()),
                ]
            )
        )

        table_def = catalog.create_table(
            table=table_name,
            namespace=self.test_namespace,
            schema=schema,
            inner=self.catalog_properties,
        )

        assert table_def.table.table_name == table_name
        assert table_def.table_version.schema.equivalent_to(schema)

        # Verify table exists
        assert catalog.table_exists(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

    def test_create_table_already_exists_fail_if_exists_true(self):
        """Test creating a table that already exists with fail_if_exists=True"""
        table_name = "test_create_table_exists"
        schema = Schema.of(schema=pa.schema([("id", pa.int64())]))

        # Create table first
        catalog.create_table(
            table=table_name,
            namespace=self.test_namespace,
            schema=schema,
            inner=self.catalog_properties,
        )

        # Try to create again with fail_if_exists=True (default)
        with pytest.raises(TableAlreadyExistsError):
            catalog.create_table(
                table=table_name,
                namespace=self.test_namespace,
                schema=schema,
                fail_if_exists=True,
                inner=self.catalog_properties,
            )

    def test_create_table_already_exists_fail_if_exists_false(self):
        """Test creating a table that already exists with fail_if_exists=False"""
        table_name = "test_create_table_exists_ok"
        schema = Schema.of(schema=pa.schema([("id", pa.int64())]))

        # Create table first
        table_def1 = catalog.create_table(
            table=table_name,
            namespace=self.test_namespace,
            schema=schema,
            inner=self.catalog_properties,
        )

        # Create again with fail_if_exists=False should return existing table
        table_def2 = catalog.create_table(
            table=table_name,
            namespace=self.test_namespace,
            schema=schema,
            fail_if_exists=False,
            inner=self.catalog_properties,
        )

        assert table_def1.table.table_name == table_def2.table.table_name


class TestWriteToTable:
    """Test the write_to_table implementation with different modes and data types."""

    @classmethod
    def setup_class(cls):
        cls.temp_dir = tempfile.mkdtemp()
        cls.catalog_properties = get_catalog_properties(root=cls.temp_dir)

        # Create a test namespace
        cls.test_namespace = "test_write_to_table"
        catalog.create_namespace(
            namespace=cls.test_namespace, inner=cls.catalog_properties
        )

    @classmethod
    def teardown_class(cls):
        shutil.rmtree(cls.temp_dir)

    def _create_test_pandas_data(self):
        """Create test pandas DataFrame"""
        return pd.DataFrame(
            {
                "id": [1, 2, 3, 4, 5],
                "name": ["Alice", "Bob", "Charlie", "Dave", "Eve"],
                "age": [25, 30, 35, 40, 45],
                "city": ["NYC", "LA", "Chicago", "Houston", "Phoenix"],
            }
        )

    def _create_test_pyarrow_data(self):
        """Create test PyArrow Table"""
        return pa.table(
            {
                "id": [1, 2, 3, 4, 5],
                "name": ["Alice", "Bob", "Charlie", "Dave", "Eve"],
                "age": [25, 30, 35, 40, 45],
                "city": ["NYC", "LA", "Chicago", "Houston", "Phoenix"],
            }
        )

    def _create_test_polars_data(self):
        """Create test Polars DataFrame"""
        return pl.DataFrame(
            {
                "id": [1, 2, 3, 4, 5],
                "name": ["Alice", "Bob", "Charlie", "Dave", "Eve"],
                "age": [25, 30, 35, 40, 45],
                "city": ["NYC", "LA", "Chicago", "Houston", "Phoenix"],
            }
        )

    def _create_second_batch_pandas_data(self):
        """Create second batch of test data for append tests"""
        return pd.DataFrame(
            {
                "id": [6, 7, 8],
                "name": ["Frank", "Grace", "Henry"],
                "age": [50, 55, 60],
                "city": ["Boston", "Seattle", "Denver"],
            }
        )

    def _create_test_ray_data(self):
        """Create test Ray Dataset for schema inference testing."""
        import ray

        # Initialize Ray if not already initialized
        # Note: Use distributed mode (not local_mode=True) to avoid Ray 2.46.0 internal bug
        if not ray.is_initialized():
            ray.init()

        data = pa.table(
            {
                "id": [1, 2, 3, 4, 5],
                "name": ["Alice", "Bob", "Charlie", "Dave", "Eve"],
                "age": [25, 30, 35, 40, 45],
                "city": ["NYC", "LA", "Chicago", "Houston", "Phoenix"],
            }
        )
        return rd.from_arrow(data)

    def _create_test_daft_data(self):
        """Create test Daft DataFrame for schema inference testing."""
        data = {
            "id": [1, 2, 3],
            "name": ["Alice", "Bob", "Charlie"],
            "age": [25, 30, 35],
            "city": ["NYC", "LA", "Chicago"],
        }
        return daft.from_pydict(data)

    def _create_test_numpy_1d_data(self):
        """Create test 1D numpy array for schema inference testing."""
        return np.array([1, 2, 3, 4, 5])

    def _create_test_numpy_2d_data(self):
        """Create test 2D numpy array for schema inference testing."""
        return np.array([[1, 25], [2, 30], [3, 35]], dtype=np.int64)

    def _create_table_with_merge_keys(self, table_name: str):
        """Create a table with merge keys for testing MERGE mode"""
        from deltacat.storage.model.schema import Schema, Field

        # Create schema with merge keys
        schema = Schema.of(
            [
                Field.of(pa.field("id", pa.int64()), is_merge_key=True),  # merge key
                Field.of(pa.field("name", pa.string())),
                Field.of(pa.field("age", pa.int32())),
                Field.of(pa.field("city", pa.string())),
            ]
        )

        catalog.create_table(
            table=table_name,
            namespace=self.test_namespace,
            schema=schema,
            inner=self.catalog_properties,
        )

        return schema

    def _create_table_without_merge_keys(self, table_name: str):
        """Create a table without merge keys for testing APPEND mode"""
        # Use schema inference with no merge keys
        data = self._create_test_pandas_data()
        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

    # Test TableWriteMode.AUTO
    def test_write_to_table_auto_create_new_table_pandas(self):
        """Test AUTO mode creating a new table with pandas data"""
        table_name = "test_auto_create_pandas"
        data = self._create_test_pandas_data()

        # Table doesn't exist, AUTO should create it
        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.AUTO,
            inner=self.catalog_properties,
        )

        # Verify table was created
        assert catalog.table_exists(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        # Verify table has correct schema
        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )
        assert table_def.table_version.schema is not None

    def test_write_to_table_auto_create_new_table_pyarrow(self):
        """Test AUTO mode creating a new table with PyArrow data"""
        table_name = "test_auto_create_pyarrow"
        data = self._create_test_pyarrow_data()

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.AUTO,
            inner=self.catalog_properties,
        )

        assert catalog.table_exists(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

    def test_write_to_table_auto_create_new_table_polars(self):
        """Test AUTO mode creating a new table with Polars data"""
        table_name = "test_auto_create_polars"
        data = self._create_test_polars_data()

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.AUTO,
            inner=self.catalog_properties,
        )

        assert catalog.table_exists(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

    def test_write_to_table_auto_append_existing_table(self):
        """Test AUTO mode appending to existing table"""
        table_name = "test_auto_append"
        data1 = self._create_test_pandas_data()
        data2 = self._create_second_batch_pandas_data()

        # First write creates table
        catalog.write_to_table(
            data=data1,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.AUTO,
            inner=self.catalog_properties,
        )

        # Second write should append
        catalog.write_to_table(
            data=data2,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.AUTO,
            inner=self.catalog_properties,
        )

        # Verify table still exists
        assert catalog.table_exists(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

    # Test TableWriteMode.CREATE
    def test_write_to_table_create_new_table(self):
        """Test CREATE mode with new table"""
        table_name = "test_create_new"
        data = self._create_test_pandas_data()

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        assert catalog.table_exists(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

    def test_write_to_table_create_existing_table_fails(self):
        """Test CREATE mode fails when table exists"""
        table_name = "test_create_fail"
        data = self._create_test_pandas_data()

        # Create table first
        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        # Try to create again should fail
        with pytest.raises(
            TableAlreadyExistsError, match="already exists and mode is CREATE"
        ):
            catalog.write_to_table(
                data=data,
                table=table_name,
                namespace=self.test_namespace,
                mode=TableWriteMode.CREATE,
                inner=self.catalog_properties,
            )

    # Test TableWriteMode.APPEND
    def test_write_to_table_append_existing_table(self):
        """Test APPEND mode with existing table"""
        table_name = "test_append_existing"
        data1 = self._create_test_pandas_data()
        data2 = self._create_second_batch_pandas_data()

        # Create table first
        catalog.write_to_table(
            data=data1,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        # Append to existing table
        catalog.write_to_table(
            data=data2,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.APPEND,
            inner=self.catalog_properties,
        )

    def test_write_to_table_append_nonexistent_table_fails(self):
        """Test APPEND mode fails when table doesn't exist"""
        table_name = "test_append_fail"
        data = self._create_test_pandas_data()

        with pytest.raises(
            TableNotFoundError,
            match="does not exist and write mode is append. Use CREATE or AUTO mode",
        ):
            catalog.write_to_table(
                data=data,
                table=table_name,
                namespace=self.test_namespace,
                mode=TableWriteMode.APPEND,
                inner=self.catalog_properties,
            )

    def test_write_to_table_append_with_merge_keys_fails(self):
        """Test APPEND mode fails when table has merge keys"""
        table_name = "test_append_with_merge_keys"

        # Create a table with merge keys
        self._create_table_with_merge_keys(table_name)

        # Create test data that matches the schema
        data = pd.DataFrame(
            {
                "id": [1, 2, 3],
                "name": ["Alice", "Bob", "Charlie"],
                "age": [25, 30, 35],
                "city": ["NYC", "LA", "Chicago"],
            }
        )

        # APPEND mode should fail since table has merge keys
        with pytest.raises(
            SchemaValidationError,
            match="APPEND mode cannot be used with tables that have merge keys",
        ):
            catalog.write_to_table(
                data=data,
                table=table_name,
                namespace=self.test_namespace,
                mode=TableWriteMode.APPEND,
                inner=self.catalog_properties,
            )

    def test_write_to_table_append_without_merge_keys_succeeds(self):
        """Test APPEND mode works when table has no merge keys"""
        table_name = "test_append_no_merge_keys"

        # Create a table without merge keys
        self._create_table_without_merge_keys(table_name)

        # Add more data to the table
        data2 = self._create_second_batch_pandas_data()

        # APPEND mode should work since table has no merge keys
        catalog.write_to_table(
            data=data2,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.APPEND,
            inner=self.catalog_properties,
        )

        # Table should still exist
        assert catalog.table_exists(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

    # Test explicit schema specification
    def test_write_to_table_explicit_schema(self):
        """Test writing with explicit schema specification"""
        table_name = "test_explicit_schema"
        data = self._create_test_pandas_data()

        # Define explicit schema with COERCE consistency types to preserve exact types
        explicit_schema = Schema.of(
            schema=[
                Field.of(
                    pa.field("id", pa.int64()),
                    consistency_type=SchemaConsistencyType.COERCE,
                ),
                Field.of(
                    pa.field("name", pa.string()),
                    consistency_type=SchemaConsistencyType.COERCE,
                ),
                Field.of(
                    pa.field("age", pa.int32()),
                    consistency_type=SchemaConsistencyType.COERCE,
                ),  # Different from inferred schema
                Field.of(
                    pa.field("city", pa.string()),
                    consistency_type=SchemaConsistencyType.COERCE,
                ),
            ]
        )

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            schema=explicit_schema,
            inner=self.catalog_properties,
        )

        # Verify schema was used
        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )
        assert table_def.table_version.schema.equivalent_to(explicit_schema)

    def test_write_to_table_explicit_schema_none(self):
        """Test writing with explicit schema=None to create schemaless table"""
        table_name = "test_explicit_schema_none"
        data = self._create_test_pandas_data()

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            schema=None,  # Explicitly set schema=None
            inner=self.catalog_properties,
        )

        # Verify table was created with schema=None (schemaless)
        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        # The table should exist but have a None/empty schema
        assert table_def is not None
        # Note: The exact behavior of schemaless tables may vary by storage implementation
        # We're mainly testing that the create_table call succeeded with schema=None

    def test_schema_behavior_comparison(self):
        """Test that demonstrates the difference between no schema vs explicit schema=None"""
        data = self._create_test_pandas_data()

        # Case 1: No schema argument - should infer schema
        table_name_inferred = "test_schema_inferred"
        catalog.write_to_table(
            data=data,
            table=table_name_inferred,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            # No schema argument provided - should infer from data
            inner=self.catalog_properties,
        )

        # Case 2: Explicit schema=None - should create schemaless table
        table_name_schemaless = "test_schema_none"
        catalog.write_to_table(
            data=data,
            table=table_name_schemaless,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            schema=None,  # Explicitly set schema=None
            inner=self.catalog_properties,
        )

        # Verify both tables were created
        table_inferred = catalog.get_table(
            table=table_name_inferred,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        table_schemaless = catalog.get_table(
            table=table_name_schemaless,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        # Both tables should exist
        assert table_inferred is not None
        assert table_schemaless is not None

        # The inferred schema table should have a schema with the expected columns
        inferred_schema = table_inferred.table_version.schema.arrow
        assert "id" in inferred_schema.names
        assert "name" in inferred_schema.names
        assert "age" in inferred_schema.names
        assert "city" in inferred_schema.names

    # Test schema inference from different data types
    def test_schema_inference_pandas(self):
        """Test schema inference from pandas DataFrame"""
        table_name = "test_schema_inference_pandas"
        data = pd.DataFrame(
            {
                "int_col": [1, 2, 3],
                "float_col": [1.1, 2.2, 3.3],
                "str_col": ["a", "b", "c"],
                "bool_col": [True, False, True],
            }
        )

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        schema = table_def.table_version.schema.arrow
        assert "int_col" in schema.names
        assert "float_col" in schema.names
        assert "str_col" in schema.names
        assert "bool_col" in schema.names

    def test_schema_inference_pyarrow(self):
        """Test schema inference from PyArrow Table"""
        table_name = "test_schema_inference_pyarrow"
        data = pa.table(
            {
                "int64_col": pa.array([1, 2, 3], type=pa.int64()),
                "string_col": pa.array(["x", "y", "z"], type=pa.string()),
                "double_col": pa.array([1.1, 2.2, 3.3], type=pa.float64()),
            }
        )

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        schema = table_def.table_version.schema.arrow
        assert schema.field("int64_col").type == pa.int64()
        assert schema.field("string_col").type == pa.string()
        assert schema.field("double_col").type == pa.float64()

    def test_schema_inference_polars(self):
        """Test schema inference from Polars DataFrame"""
        table_name = "test_schema_inference_polars"
        data = pl.DataFrame(
            {
                "int_col": [1, 2, 3],
                "str_col": ["a", "b", "c"],
                "float_col": [1.1, 2.2, 3.3],
            }
        )

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        schema = table_def.table_version.schema.arrow
        assert "int_col" in schema.names
        assert "str_col" in schema.names
        assert "float_col" in schema.names

    def test_schema_inference_ray_dataset(self):
        """Test schema inference from Ray Dataset"""
        table_name = "test_schema_inference_ray"
        ray_data = self._create_test_ray_data()

        catalog.write_to_table(
            data=ray_data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        schema = table_def.table_version.schema.arrow
        assert "id" in schema.names
        assert "name" in schema.names
        assert "age" in schema.names
        assert "city" in schema.names

    def test_schema_inference_daft_dataframe(self):
        """Test schema inference from Daft DataFrame"""
        table_name = "test_schema_inference_daft"
        data = self._create_test_daft_data()

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        schema = table_def.table_version.schema.arrow
        assert "id" in schema.names
        assert "name" in schema.names
        assert "age" in schema.names
        assert "city" in schema.names

    def test_schema_inference_numpy_1d(self):
        """Test schema inference from 1D numpy array"""
        table_name = "test_schema_inference_numpy_1d"
        data = self._create_test_numpy_1d_data()

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        schema = table_def.table_version.schema.arrow
        assert (
            "0" in schema.names
        )  # pandas converts 1D numpy array with column name "0"
        assert len(schema.names) == 1

    def test_schema_inference_numpy_2d(self):
        """Test schema inference from 2D numpy array"""
        table_name = "test_schema_inference_numpy_2d"
        data = self._create_test_numpy_2d_data()

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        schema = table_def.table_version.schema.arrow
        assert (
            "0" in schema.names
        )  # pandas converts 2D numpy array with column names "0", "1"
        assert "1" in schema.names
        assert len(schema.names) == 2

    def test_numpy_3d_array_error(self):
        """Test that 3D numpy arrays raise an error"""
        table_name = "test_numpy_3d_error"
        data = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])  # 3D array

        with pytest.raises(
            ValueError, match="NumPy arrays with 3 dimensions are not supported"
        ):
            catalog.write_to_table(
                data=data,
                table=table_name,
                namespace=self.test_namespace,
                mode=TableWriteMode.CREATE,
                inner=self.catalog_properties,
            )

    # Test different content types
    def test_write_to_table_different_content_types(self):
        """Test writing with different content types"""
        data = self._create_test_pandas_data()

        content_types = [
            ContentType.PARQUET,
            ContentType.CSV,
            ContentType.JSON,
        ]

        for i, content_type in enumerate(content_types):
            table_name = f"test_content_type_{content_type.value}_{i}"

            catalog.write_to_table(
                data=data,
                table=table_name,
                namespace=self.test_namespace,
                mode=TableWriteMode.CREATE,
                content_type=content_type,
                inner=self.catalog_properties,
                schema=None,
            )

            assert catalog.table_exists(
                table=table_name,
                namespace=self.test_namespace,
                inner=self.catalog_properties,
            )

    # Test table creation parameters
    def test_write_to_table_with_table_properties(self):
        """Test writing with table creation parameters"""
        table_name = "test_table_properties"
        data = self._create_test_pandas_data()

        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            table_description="Test table with properties",
            lifecycle_state=LifecycleState.ACTIVE,
            inner=self.catalog_properties,
        )

        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        assert table_def.table.description == "Test table with properties"
        # Note: lifecycle_state defaults to ACTIVE in create_table, but may be overridden
        # We'll accept either ACTIVE or CREATED as both are valid for our test purpose
        assert table_def.table_version.state in [
            LifecycleState.ACTIVE,
            LifecycleState.CREATED,
        ]

    # Test error conditions
    def test_write_to_table_unsupported_data_type(self):
        """Test error when data type cannot be inferred"""
        table_name = "test_unsupported_data"

        # Use a plain dict which doesn't have schema inference
        unsupported_data = {"key": "value"}

        with pytest.raises(
            ValueError, match="No schema inference function found for table type"
        ):
            catalog.write_to_table(
                data=unsupported_data,
                table=table_name,
                namespace=self.test_namespace,
                mode=TableWriteMode.CREATE,
                inner=self.catalog_properties,
            )

    def test_write_to_table_replace_mode(self):
        """Test REPLACE mode creating a new stream to replace existing data"""
        table_name = "test_replace_mode"
        data1 = self._create_test_pandas_data()
        data2 = self._create_second_batch_pandas_data()

        # First, create the table
        catalog.write_to_table(
            data=data1,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        # Verify table exists
        assert catalog.table_exists(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        # Now use REPLACE mode to replace all existing data
        catalog.write_to_table(
            data=data2,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.REPLACE,
            inner=self.catalog_properties,
        )

        # Table should still exist
        assert catalog.table_exists(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

    def test_write_to_table_merge_mode_with_merge_keys(self):
        """Test MERGE mode works when table has merge keys"""
        table_name = "test_merge_mode_with_keys"

        # Create a table with merge keys
        self._create_table_with_merge_keys(table_name)

        # Create test data that matches the schema
        data = pd.DataFrame(
            {
                "id": [1, 2, 3],
                "name": ["Alice", "Bob", "Charlie"],
                "age": [25, 30, 35],
                "city": ["NYC", "LA", "Chicago"],
            }
        )

        # MERGE mode should work since table has merge keys
        catalog.write_to_table(
            data=data,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.MERGE,
            inner=self.catalog_properties,
        )

        # Table should still exist
        assert catalog.table_exists(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

    def test_write_to_table_merge_mode_without_merge_keys_fails(self):
        """Test MERGE mode fails when table has no merge keys"""
        table_name = "test_merge_mode_no_keys"

        # Create a table without merge keys
        self._create_table_without_merge_keys(table_name)

        data = self._create_test_pandas_data()

        # MERGE mode should fail since table has no merge keys
        with pytest.raises(
            TableValidationError,
            match="MERGE mode requires tables to have at least one merge key",
        ):
            catalog.write_to_table(
                data=data,
                table=table_name,
                namespace=self.test_namespace,
                mode=TableWriteMode.MERGE,
                inner=self.catalog_properties,
            )

    # Test default namespace behavior
    def test_write_to_table_default_namespace(self):
        """Test writing to table using default namespace"""
        table_name = "test_default_namespace"
        data = self._create_test_pandas_data()

        # Don't specify namespace, should use default
        catalog.write_to_table(
            data=data,
            table=table_name,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        # Should be able to find table in default namespace
        default_ns = catalog.default_namespace(inner=self.catalog_properties)
        assert catalog.table_exists(
            table=table_name, namespace=default_ns, inner=self.catalog_properties
        )

    def test_write_to_table_append_creates_separate_deltas(self):
        """Test that APPEND mode creates separate deltas in the same partition"""
        from deltacat.catalog.main.impl import _get_storage

        table_name = "test_append_separate_deltas"
        data1 = self._create_test_pandas_data()
        data2 = self._create_second_batch_pandas_data()

        # Create table with first batch
        catalog.write_to_table(
            data=data1,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.CREATE,
            inner=self.catalog_properties,
        )

        # Get the table definition to access stream information
        table_def = catalog.get_table(
            table=table_name,
            namespace=self.test_namespace,
            inner=self.catalog_properties,
        )

        # Get storage interface
        storage = _get_storage(inner=self.catalog_properties)

        # Get the stream
        stream = storage.get_stream(
            namespace=self.test_namespace,
            table_name=table_name,
            table_version=table_def.table_version.table_version,
            inner=self.catalog_properties,
        )

        # Get the partition (should be only one for unpartitioned table)
        partition = storage.get_partition(
            stream_locator=stream.locator,
            partition_values=None,  # unpartitioned
            inner=self.catalog_properties,
        )

        # List deltas before second write
        deltas_before = storage.list_partition_deltas(
            partition_like=partition,
            inner=self.catalog_properties,
        ).all_items()

        assert (
            len(deltas_before) == 1
        ), f"Expected 1 delta before append, got {len(deltas_before)}"

        # Append second batch using APPEND mode
        catalog.write_to_table(
            data=data2,
            table=table_name,
            namespace=self.test_namespace,
            mode=TableWriteMode.APPEND,
            inner=self.catalog_properties,
        )

        # Get the same partition again (should be the same partition object)
        partition_after = storage.get_partition(
            stream_locator=stream.locator,
            partition_values=None,  # unpartitioned
            inner=self.catalog_properties,
        )

        # Verify it's the same partition
        assert (
            partition.partition_id == partition_after.partition_id
        ), "APPEND should reuse the same partition"

        # List deltas after second write
        deltas_after = storage.list_partition_deltas(
            partition_like=partition_after,
            inner=self.catalog_properties,
        ).all_items()

        # Should now have 2 deltas in the same partition
        assert (
            len(deltas_after) == 2
        ), f"Expected 2 deltas after append, got {len(deltas_after)}"

        # Verify deltas have different stream positions
        stream_positions = [delta.stream_position for delta in deltas_after]
        assert (
            len(set(stream_positions)) == 2
        ), "Deltas should have different stream positions"
        assert min(stream_positions) == 1, "First delta should have stream position 1"
        assert max(stream_positions) == 2, "Second delta should have stream position 2"

    def test_write_to_table_partitioned_table_raises_not_implemented(self):
        """Test that write_to_table raises NotImplementedError for partitioned tables"""
        from deltacat.storage.model.partition import (
            PartitionScheme,
            PartitionKey,
            PartitionKeyList,
        )
        from deltacat.storage.model.transform import IdentityTransform

        table_name = "test_partitioned_table"
        data = self._create_test_pandas_data()

        # Create a partition scheme with partition keys
        partition_keys = [
            PartitionKey.of(
                key=["city"],
                name="city_partition",
                transform=IdentityTransform.of(),
            )
        ]
        partition_scheme = PartitionScheme.of(
            keys=PartitionKeyList.of(partition_keys),
            name="test_partition_scheme",
            scheme_id="test_partition_scheme_id",
        )

        # Try to create a partitioned table using write_to_table
        with pytest.raises(
            NotImplementedError,
            match="write_to_table does not yet support partitioned tables",
        ):
            catalog.write_to_table(
                data=data,
                table=table_name,
                namespace=self.test_namespace,
                mode=TableWriteMode.CREATE,
                partition_scheme=partition_scheme,  # This makes it partitioned
                inner=self.catalog_properties,
            )

    def test_write_to_table_sorted_table_raises_not_implemented(self):
        """Test that write_to_table raises NotImplementedError for tables with sort keys"""
        from deltacat.storage.model.sort_key import SortScheme, SortKey, SortKeyList
        from deltacat.storage.model.types import SortOrder, NullOrder

        table_name = "test_sorted_table"
        data = self._create_test_pandas_data()

        # Create sort scheme with sort keys
        sort_scheme = SortScheme.of(
            keys=SortKeyList.of(
                [
                    SortKey.of(
                        key=["id"],
                        sort_order=SortOrder.ASCENDING,
                        null_order=NullOrder.AT_END,
                    )
                ]
            ),
            name="test_sort_scheme",
            scheme_id="test_sort_scheme_id",
        )

        # Create table with sort keys
        catalog.create_table(
            table=table_name,
            namespace=self.test_namespace,
            sort_keys=sort_scheme,
            inner=self.catalog_properties,
        )

        # Attempt to write to the sorted table should raise NotImplementedError
        with pytest.raises(NotImplementedError) as exc_info:
            catalog.write_to_table(
                data=data,
                table=table_name,
                namespace=self.test_namespace,
                mode=TableWriteMode.APPEND,
                inner=self.catalog_properties,
            )

        # Verify the error message contains expected information
        assert "sort keys" in str(exc_info.value)
        assert "sort scheme with 1 sort key(s)" in str(exc_info.value)
        assert "id" in str(exc_info.value)



from unittest import TestCase
import polars as pl
import pandas as pd
import tempfile
import fsspec
import gzip
import json
import io
from deltacat.types.media import ContentType, ContentEncoding
from deltacat.utils.polars import (
    dataframe_to_file,
    file_to_dataframe,
    content_type_to_reader_kwargs,
    _add_column_kwargs,
    ReadKwargsProviderPolarsStringTypes,
    concat_dataframes,
)


class TestPolarsWriters(TestCase):
    def setUp(self):
        # Create a test DataFrame with data that includes delimiters
        self.df = pl.DataFrame({"col1": ["a,b\tc|d", "e,f\tg|h"], "col2": [1, 2]})
        self.fs = fsspec.filesystem("file")
        self.base_path = tempfile.mkdtemp()
        self.fs.makedirs(self.base_path, exist_ok=True)

    def tearDown(self):
        self.fs.rm(self.base_path, recursive=True)

    def test_write_feather(self):
        path = f"{self.base_path}/test.feather"

        dataframe_to_file(
            self.df,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.FEATHER.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content
        result = pl.read_ipc(path)
        assert result.equals(self.df)

    def test_write_csv(self):
        path = f"{self.base_path}/test.csv.gz"

        dataframe_to_file(
            self.df, path, self.fs, lambda x: path, content_type=ContentType.CSV.value
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content (should be GZIP compressed)
        with self.fs.open(path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                content = gz.read().decode("utf-8")
                # Should be quoted due to commas in data
                assert '"a,b\tc|d",1' in content
                assert '"e,f\tg|h",2' in content

    def test_write_tsv(self):
        path = f"{self.base_path}/test.tsv.gz"

        dataframe_to_file(
            self.df,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.TSV.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content (should be GZIP compressed)
        with self.fs.open(path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                content = gz.read().decode("utf-8")
                # Polars writes TSV with tab separators
                assert '"a,b\tc|d"\t1' in content
                assert '"e,f\tg|h"\t2' in content

    def test_write_psv(self):
        path = f"{self.base_path}/test.psv.gz"

        dataframe_to_file(
            self.df,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.PSV.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content (should be GZIP compressed)
        with self.fs.open(path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                content = gz.read().decode("utf-8")
                # Polars writes PSV with pipe separators
                assert '"a,b\tc|d"|1' in content
                assert '"e,f\tg|h"|2' in content

    def test_write_unescaped_tsv(self):
        # Create DataFrame without delimiters for unescaped TSV
        df = pl.DataFrame({"col1": ["abc", "def"], "col2": [1, 2]})
        path = f"{self.base_path}/test.tsv.gz"

        dataframe_to_file(
            df,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.UNESCAPED_TSV.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content (should be GZIP compressed)
        with self.fs.open(path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                content = gz.read().decode("utf-8")
                # With quote_char=None for unescaped TSV, should use tab separators
                assert "abc\t1" in content
                assert "def\t2" in content

    def test_write_orc(self):
        path = f"{self.base_path}/test.orc"

        dataframe_to_file(
            self.df, path, self.fs, lambda x: path, content_type=ContentType.ORC.value
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content by reading with pandas (since polars delegates to PyArrow)
        result = pd.read_orc(path)
        expected = self.df.to_pandas()
        pd.testing.assert_frame_equal(result, expected)

    def test_write_parquet(self):
        path = f"{self.base_path}/test.parquet"

        dataframe_to_file(
            self.df,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.PARQUET.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content
        result = pl.read_parquet(path)
        assert result.equals(self.df)

    def test_write_json(self):
        path = f"{self.base_path}/test.json.gz"

        dataframe_to_file(
            self.df, path, self.fs, lambda x: path, content_type=ContentType.JSON.value
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content (should be GZIP compressed, newline-delimited JSON)
        with self.fs.open(path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                content = gz.read().decode("utf-8")
                # Each line should be a valid JSON object
                lines = [
                    line for line in content.split("\n") if line
                ]  # Skip empty lines
                assert len(lines) == 2  # 2 records
                assert json.loads(lines[0]) == {"col1": "a,b\tc|d", "col2": 1}
                assert json.loads(lines[1]) == {"col1": "e,f\tg|h", "col2": 2}

    def test_write_avro(self):
        path = f"{self.base_path}/test.avro"

        dataframe_to_file(
            self.df, path, self.fs, lambda x: path, content_type=ContentType.AVRO.value
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content by reading with polars
        result = pl.read_avro(path)
        assert result.equals(self.df)


class TestPolarsReaders(TestCase):
    def setUp(self):
        # Create test data files for reading
        self.fs = fsspec.filesystem("file")
        self.base_path = tempfile.mkdtemp()
        self.fs.makedirs(self.base_path, exist_ok=True)

        # Create test DataFrame
        self.df = pl.DataFrame(
            {
                "col1": ["a,b\tc|d", "e,f\tg|h", "test"],
                "col2": [1, 2, 3],
                "col3": [1.1, 2.2, 3.3],
            }
        )

        # Write test files in different formats
        self._create_test_files()

    def tearDown(self):
        self.fs.rm(self.base_path, recursive=True)

    def _create_test_files(self):
        """Create test files for reading tests with the original test data structure."""
        import gzip
        import bz2

        # Create CSV file (GZIP compressed) with the original test data
        csv_path = f"{self.base_path}/test.csv"
        with self.fs.open(csv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = '"a,b\tc|d",1,1.1\n"e,f\tg|h",2,2.2\ntest,3,3.3\n'
                gz.write(content.encode("utf-8"))

        # Create TSV file (GZIP compressed)
        tsv_path = f"{self.base_path}/test.tsv"
        with self.fs.open(tsv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = '"a,b\tc|d"\t1\t1.1\n"e,f\tg|h"\t2\t2.2\ntest\t3\t3.3\n'
                gz.write(content.encode("utf-8"))

        # Create PSV file (GZIP compressed)
        psv_path = f"{self.base_path}/test.psv"
        with self.fs.open(psv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = '"a,b\tc|d"|1|1.1\n"e,f\tg|h"|2|2.2\ntest|3|3.3\n'
                gz.write(content.encode("utf-8"))

        # Create unescaped TSV file (GZIP compressed)
        unescaped_tsv_path = f"{self.base_path}/test_unescaped.tsv"
        with self.fs.open(unescaped_tsv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = "abc\t1\t1.1\ndef\t2\t2.2\nghi\t3\t3.3\n"
                gz.write(content.encode("utf-8"))

        # Create Parquet file
        parquet_path = f"{self.base_path}/test.parquet"
        self.df.write_parquet(parquet_path)

        # Create Feather file
        feather_path = f"{self.base_path}/test.feather"
        self.df.write_ipc(feather_path)

        # Create JSON file (GZIP compressed, NDJSON format)
        json_path = f"{self.base_path}/test.json"
        with self.fs.open(json_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                # Use proper NDJSON format - one JSON object per line
                lines = []
                for i in range(len(self.df)):
                    row = self.df.row(i)
                    json_obj = {"col1": row[0], "col2": row[1], "col3": row[2]}
                    lines.append(json.dumps(json_obj))
                content = "\n".join(lines) + "\n"
                gz.write(content.encode("utf-8"))

        # Create Avro file
        avro_path = f"{self.base_path}/test.avro"
        self.df.write_avro(avro_path)

        # Create ORC file using pandas (since polars delegates to pandas for ORC)
        orc_path = f"{self.base_path}/test.orc"
        self.df.to_pandas().to_orc(orc_path)

        # Create BZIP2 compressed CSV for compression tests
        bzip2_path = f"{self.base_path}/test_bzip2.csv.bz2"
        with bz2.open(bzip2_path, "wt") as f:
            f.write('"a,b\tc|d",1,1.1\n"e,f\tg|h",2,2.2\ntest,3,3.3\n')

    def test_content_type_to_reader_kwargs(self):
        # Test CSV kwargs
        csv_kwargs = content_type_to_reader_kwargs(ContentType.CSV.value)
        expected_csv = {"separator": ",", "has_header": False}
        assert csv_kwargs == expected_csv

        # Test TSV kwargs
        tsv_kwargs = content_type_to_reader_kwargs(ContentType.TSV.value)
        expected_tsv = {"separator": "\t", "has_header": False}
        assert tsv_kwargs == expected_tsv

        # Test PSV kwargs
        psv_kwargs = content_type_to_reader_kwargs(ContentType.PSV.value)
        expected_psv = {"separator": "|", "has_header": False}
        assert psv_kwargs == expected_psv

        # Test unescaped TSV kwargs
        unescaped_kwargs = content_type_to_reader_kwargs(
            ContentType.UNESCAPED_TSV.value
        )
        expected_unescaped = {
            "separator": "\t",
            "has_header": False,
            "null_values": [""],
            "quote_char": None,
        }
        assert unescaped_kwargs == expected_unescaped

        # Test Parquet kwargs (should be empty)
        parquet_kwargs = content_type_to_reader_kwargs(ContentType.PARQUET.value)
        assert parquet_kwargs == {}

    def test_add_column_kwargs(self):
        kwargs = {}
        column_names = ["col1", "col2", "col3"]
        include_columns = ["col1", "col2"]

        # Test CSV column kwargs
        _add_column_kwargs(ContentType.CSV.value, column_names, include_columns, kwargs)
        assert kwargs["new_columns"] == column_names
        assert kwargs["columns"] == include_columns

        # Test Parquet column kwargs
        kwargs = {}
        _add_column_kwargs(
            ContentType.PARQUET.value, column_names, include_columns, kwargs
        )
        assert kwargs["columns"] == include_columns
        assert "new_columns" not in kwargs

    def test_read_csv_from_file(self):
        csv_path = f"{self.base_path}/test.csv"

        # Read using polars directly to test our reader logic
        with self.fs.open(csv_path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                source = io.BytesIO(gz.read())

        kwargs = content_type_to_reader_kwargs(ContentType.CSV.value)
        kwargs["new_columns"] = ["col1", "col2", "col3"]

        result = pl.read_csv(source, **kwargs)

        # Verify basic structure
        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result["col1"].to_list() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_read_tsv_from_file(self):
        tsv_path = f"{self.base_path}/test.tsv"

        with self.fs.open(tsv_path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                source = io.BytesIO(gz.read())

        kwargs = content_type_to_reader_kwargs(ContentType.TSV.value)
        kwargs["new_columns"] = ["col1", "col2", "col3"]

        result = pl.read_csv(source, **kwargs)

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result["col1"].to_list() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_read_psv_from_file(self):
        psv_path = f"{self.base_path}/test.psv"

        with self.fs.open(psv_path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                source = io.BytesIO(gz.read())

        kwargs = content_type_to_reader_kwargs(ContentType.PSV.value)
        kwargs["new_columns"] = ["col1", "col2", "col3"]

        result = pl.read_csv(source, **kwargs)

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result["col1"].to_list() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_read_parquet_from_file(self):
        parquet_path = f"{self.base_path}/test.parquet"
        result = pl.read_parquet(parquet_path)

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result.equals(self.df)

    def test_read_feather_from_file(self):
        feather_path = f"{self.base_path}/test.feather"
        result = pl.read_ipc(feather_path)

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result.equals(self.df)

    def test_read_json_from_file(self):
        json_path = f"{self.base_path}/test.json"

        with self.fs.open(json_path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                source = io.BytesIO(gz.read())

        result = pl.read_ndjson(source)

        assert len(result) == 3
        assert set(result.columns) == {"col1", "col2", "col3"}
        assert result["col1"].to_list() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_read_avro_from_file(self):
        avro_path = f"{self.base_path}/test.avro"
        result = pl.read_avro(avro_path)

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result.equals(self.df)

    def test_read_orc_from_file(self):
        # Test ORC reading via pandas conversion
        orc_path = f"{self.base_path}/test.orc"

        # Read with pandas and convert to polars (mimicking our ORC handling)
        import pandas as pd

        pd_df = pd.read_orc(orc_path)
        result = pl.from_pandas(pd_df)

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        # Convert both to pandas for comparison due to potential type differences
        pd.testing.assert_frame_equal(result.to_pandas(), self.df.to_pandas())

    def test_read_kwargs_provider_string_types(self):
        # Test the string types provider
        provider = ReadKwargsProviderPolarsStringTypes()
        kwargs = {"separator": ",", "has_header": False}

        # Apply string types
        result_kwargs = provider._get_kwargs(ContentType.CSV.value, kwargs)

        # Should add infer_schema=False for string type inference
        assert "infer_schema" in result_kwargs
        assert result_kwargs["infer_schema"] is False

    def test_concat_dataframes(self):
        # Test concatenation of multiple dataframes
        df1 = pl.DataFrame({"col1": ["a"], "col2": [1]})
        df2 = pl.DataFrame({"col1": ["b"], "col2": [2]})
        df3 = pl.DataFrame({"col1": ["c"], "col2": [3]})

        # Test normal concatenation
        result = concat_dataframes([df1, df2, df3])
        assert len(result) == 3
        assert result["col1"].to_list() == ["a", "b", "c"]

        # Test single dataframe
        result = concat_dataframes([df1])
        assert result.equals(df1)

        # Test empty list
        result = concat_dataframes([])
        assert result is None

        # Test None input
        result = concat_dataframes(None)
        assert result is None

    def test_file_to_dataframe_csv(self):
        # Test reading CSV with file_to_dataframe
        csv_path = f"{self.base_path}/test.csv"

        result = file_to_dataframe(
            csv_path,
            ContentType.CSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result["col1"].to_list() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_file_to_dataframe_tsv(self):
        # Test reading TSV with file_to_dataframe
        tsv_path = f"{self.base_path}/test.tsv"

        result = file_to_dataframe(
            tsv_path,
            ContentType.TSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result["col1"].to_list() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_file_to_dataframe_parquet(self):
        # Test reading Parquet with file_to_dataframe
        parquet_path = f"{self.base_path}/test.parquet"

        result = file_to_dataframe(
            parquet_path, ContentType.PARQUET.value, filesystem=self.fs
        )

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result.equals(self.df)

    def test_file_to_dataframe_feather(self):
        # Test reading Feather with file_to_dataframe
        feather_path = f"{self.base_path}/test.feather"

        result = file_to_dataframe(
            feather_path, ContentType.FEATHER.value, filesystem=self.fs
        )

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result.equals(self.df)

    def test_file_to_dataframe_json(self):
        # Test reading JSON with file_to_dataframe
        json_path = f"{self.base_path}/test.json"

        result = file_to_dataframe(
            json_path,
            ContentType.JSON.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
        )

        assert len(result) == 3
        assert set(result.columns) == {"col1", "col2", "col3"}
        assert result["col1"].to_list() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_file_to_dataframe_avro(self):
        # Test reading Avro with file_to_dataframe
        avro_path = f"{self.base_path}/test.avro"

        result = file_to_dataframe(
            avro_path, ContentType.AVRO.value, filesystem=self.fs
        )

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result.equals(self.df)

    def test_file_to_dataframe_orc(self):
        # Test reading ORC with file_to_dataframe
        orc_path = f"{self.base_path}/test.orc"

        result = file_to_dataframe(orc_path, ContentType.ORC.value, filesystem=self.fs)

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        # Convert both to pandas for comparison due to potential type differences
        pd.testing.assert_frame_equal(result.to_pandas(), self.df.to_pandas())

    def test_file_to_dataframe_with_column_selection(self):
        # Test reading with column selection
        csv_path = f"{self.base_path}/test.csv"

        # When has_header=False and we specify columns, we need to use column indices or
        # not provide new_columns. Let's test by just specifying the first 2 columns by index
        result = file_to_dataframe(
            csv_path,
            ContentType.CSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            include_columns=[0, 1],  # Select first two columns by index
        )

        assert len(result) == 3
        assert len(result.columns) == 2  # Should only have 2 columns
        # With auto-generated column names when has_header=False
        assert list(result.columns) == ["column_1", "column_2"]

    def test_file_to_dataframe_with_kwargs_provider(self):
        # Test reading with kwargs provider
        csv_path = f"{self.base_path}/test.csv"
        provider = ReadKwargsProviderPolarsStringTypes(
            include_columns=["column_1", "column_2", "column_3"]
        )

        result = file_to_dataframe(
            csv_path,
            ContentType.CSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            pl_read_func_kwargs_provider=provider,
        )

        assert len(result) == 3
        assert list(result.columns) == ["column_1", "column_2", "column_3"]
        # With string types provider, all columns should be strings
        assert all(result[col].dtype == pl.Utf8 for col in result.columns)

    def test_file_to_dataframe_filesystem_inference(self):
        # Test filesystem inference when no filesystem is provided
        parquet_path = f"{self.base_path}/test.parquet"

        result = file_to_dataframe(
            parquet_path,
            ContentType.PARQUET.value
            # No filesystem provided - should be inferred
        )

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result.equals(self.df)

    def test_file_to_dataframe_unsupported_content_type(self):
        # Test error handling for unsupported content type
        parquet_path = f"{self.base_path}/test.parquet"

        with self.assertRaises(NotImplementedError) as context:
            file_to_dataframe(
                parquet_path, "unsupported/content-type", filesystem=self.fs
            )

        assert "not implemented" in str(context.exception)

    def test_file_to_dataframe_bzip2_compression(self):
        # Test BZIP2 compression handling
        import bz2

        # Create a BZIP2 compressed CSV file
        csv_content = '"a,b\tc|d",1,1.1\n"e,f\tg|h",2,2.2\ntest,3,3.3\n'
        compressed_content = bz2.compress(csv_content.encode("utf-8"))

        bz2_path = f"{self.base_path}/test.csv.bz2"
        with self.fs.open(bz2_path, "wb") as f:
            f.write(compressed_content)

        result = file_to_dataframe(
            bz2_path,
            ContentType.CSV.value,
            ContentEncoding.BZIP2.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert len(result) == 3
        assert list(result.columns) == ["col1", "col2", "col3"]
        assert result["col1"].to_list() == ["a,b\tc|d", "e,f\tg|h", "test"]


class TestPolarsFileSystemSupport(TestCase):
    """
    Comprehensive tests for encoding-aware reader functions with different filesystem types.
    Tests fsspec AbstractFileSystem, PyArrow FileSystem, and auto-inferred filesystem.
    """

    def setUp(self):
        import pyarrow.fs as pafs

        # Create test data
        self.test_data = pl.DataFrame(
            {
                "col1": ["value1", "value2", "value3"],
                "col2": [1, 2, 3],
                "col3": [1.1, 2.2, 3.3],
            }
        )

        # Set up temporary directory
        self.temp_dir = tempfile.mkdtemp()

        # Set up different filesystem types
        self.fsspec_fs = fsspec.filesystem("file")
        self.pyarrow_fs = pafs.LocalFileSystem()

        # Create test files for each content type
        self._create_test_files()

    def tearDown(self):
        import shutil

        shutil.rmtree(self.temp_dir)

    def _create_test_files(self):
        """Create test files in different formats with different compression types."""
        import gzip
        import bz2

        # CSV files
        csv_data = "col1,col2,col3\nvalue1,1,1.1\nvalue2,2,2.2\nvalue3,3,3.3\n"

        # Create uncompressed CSV
        with open(f"{self.temp_dir}/test.csv", "w") as f:
            f.write(csv_data)

        # Create GZIP compressed CSV (fix: properly close the file)
        with gzip.open(f"{self.temp_dir}/test_gzip.csv.gz", "wt") as f:
            f.write(csv_data)

        # Create BZIP2 compressed CSV (fix: properly close the file)
        with bz2.open(f"{self.temp_dir}/test_bzip2.csv.bz2", "wt") as f:
            f.write(csv_data)

        # Parquet file
        self.test_data.write_parquet(f"{self.temp_dir}/test.parquet")

        # Feather/IPC file
        self.test_data.write_ipc(f"{self.temp_dir}/test.feather")

        # JSON file (NDJSON)
        json_data = '{"col1":"value1","col2":1,"col3":1.1}\n{"col1":"value2","col2":2,"col3":2.2}\n{"col1":"value3","col2":3,"col3":3.3}\n'
        with open(f"{self.temp_dir}/test.json", "w") as f:
            f.write(json_data)

        # AVRO file
        self.test_data.write_avro(f"{self.temp_dir}/test.avro")

        # ORC file (via pandas since polars delegates to pandas for ORC)
        self.test_data.to_pandas().to_orc(f"{self.temp_dir}/test.orc")

    def _assert_dataframes_equal(self, result, expected):
        """Helper to assert polars dataframes are equal."""
        # Convert to pandas for comparison since polars equality can be tricky with floating point
        pd.testing.assert_frame_equal(
            result.to_pandas().reset_index(drop=True),
            expected.to_pandas().reset_index(drop=True),
            check_dtype=False,  # Allow minor type differences
        )

    def test_csv_with_fsspec_filesystem(self):
        """Test CSV reading with fsspec AbstractFileSystem."""
        from deltacat.utils.polars import read_csv

        # Test uncompressed CSV
        result = read_csv(
            f"{self.temp_dir}/test.csv",
            filesystem=self.fsspec_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
            has_header=True,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test GZIP compressed CSV
        result = read_csv(
            f"{self.temp_dir}/test_gzip.csv.gz",
            filesystem=self.fsspec_fs,
            content_encoding=ContentEncoding.GZIP.value,
            has_header=True,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test BZIP2 compressed CSV
        result = read_csv(
            f"{self.temp_dir}/test_bzip2.csv.bz2",
            filesystem=self.fsspec_fs,
            content_encoding=ContentEncoding.BZIP2.value,
            has_header=True,
        )
        self._assert_dataframes_equal(result, self.test_data)

    def test_csv_with_pyarrow_filesystem(self):
        """Test CSV reading with PyArrow FileSystem."""
        from deltacat.utils.polars import read_csv

        # Test uncompressed CSV
        result = read_csv(
            f"{self.temp_dir}/test.csv",
            filesystem=self.pyarrow_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
            has_header=True,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test GZIP compressed CSV
        result = read_csv(
            f"{self.temp_dir}/test_gzip.csv.gz",
            filesystem=self.pyarrow_fs,
            content_encoding=ContentEncoding.GZIP.value,
            has_header=True,
        )
        self._assert_dataframes_equal(result, self.test_data)

    def test_csv_with_auto_inferred_filesystem(self):
        """Test CSV reading with automatically inferred filesystem."""
        from deltacat.utils.polars import read_csv

        # Test uncompressed CSV (filesystem=None, should auto-infer)
        result = read_csv(
            f"{self.temp_dir}/test.csv",
            filesystem=None,
            content_encoding=ContentEncoding.IDENTITY.value,
            has_header=True,
        )
        self._assert_dataframes_equal(result, self.test_data)

    def test_parquet_with_different_filesystems(self):
        """Test Parquet reading with different filesystem types."""
        from deltacat.utils.polars import read_parquet

        # Test with fsspec
        result = read_parquet(
            f"{self.temp_dir}/test.parquet",
            filesystem=self.fsspec_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test with PyArrow
        result = read_parquet(
            f"{self.temp_dir}/test.parquet",
            filesystem=self.pyarrow_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test with auto-inferred
        result = read_parquet(
            f"{self.temp_dir}/test.parquet",
            filesystem=None,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

    def test_feather_with_different_filesystems(self):
        """Test Feather/IPC reading with different filesystem types."""
        from deltacat.utils.polars import read_ipc

        # Test with fsspec
        result = read_ipc(
            f"{self.temp_dir}/test.feather",
            filesystem=self.fsspec_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test with PyArrow
        result = read_ipc(
            f"{self.temp_dir}/test.feather",
            filesystem=self.pyarrow_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test with auto-inferred
        result = read_ipc(
            f"{self.temp_dir}/test.feather",
            filesystem=None,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

    def test_json_with_different_filesystems(self):
        """Test JSON reading with different filesystem types."""
        from deltacat.utils.polars import read_ndjson

        # Test with fsspec
        result = read_ndjson(
            f"{self.temp_dir}/test.json",
            filesystem=self.fsspec_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test with PyArrow
        result = read_ndjson(
            f"{self.temp_dir}/test.json",
            filesystem=self.pyarrow_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test with auto-inferred
        result = read_ndjson(
            f"{self.temp_dir}/test.json",
            filesystem=None,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

    def test_avro_with_different_filesystems(self):
        """Test AVRO reading with different filesystem types."""
        from deltacat.utils.polars import read_avro

        # Test with fsspec
        result = read_avro(
            f"{self.temp_dir}/test.avro",
            filesystem=self.fsspec_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test with PyArrow
        result = read_avro(
            f"{self.temp_dir}/test.avro",
            filesystem=self.pyarrow_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test with auto-inferred
        result = read_avro(
            f"{self.temp_dir}/test.avro",
            filesystem=None,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

    def test_orc_with_different_filesystems(self):
        """Test ORC reading with different filesystem types."""
        from deltacat.utils.polars import read_orc

        # Test with fsspec
        result = read_orc(
            f"{self.temp_dir}/test.orc",
            filesystem=self.fsspec_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test with PyArrow
        result = read_orc(
            f"{self.temp_dir}/test.orc",
            filesystem=self.pyarrow_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

        # Test with auto-inferred
        result = read_orc(
            f"{self.temp_dir}/test.orc",
            filesystem=None,
            content_encoding=ContentEncoding.IDENTITY.value,
        )
        self._assert_dataframes_equal(result, self.test_data)

    def test_file_to_dataframe_with_different_filesystems(self):
        """Test file_to_dataframe with different filesystem types for all content types."""
        test_cases = [
            (
                f"{self.temp_dir}/test.csv",
                ContentType.CSV.value,
                ContentEncoding.IDENTITY.value,
                {"has_header": True},
            ),
            (
                f"{self.temp_dir}/test_gzip.csv.gz",
                ContentType.CSV.value,
                ContentEncoding.GZIP.value,
                {"has_header": True},
            ),
            (
                f"{self.temp_dir}/test.parquet",
                ContentType.PARQUET.value,
                ContentEncoding.IDENTITY.value,
                {},
            ),
            (
                f"{self.temp_dir}/test.feather",
                ContentType.FEATHER.value,
                ContentEncoding.IDENTITY.value,
                {},
            ),
            (
                f"{self.temp_dir}/test.json",
                ContentType.JSON.value,
                ContentEncoding.IDENTITY.value,
                {},
            ),
            (
                f"{self.temp_dir}/test.avro",
                ContentType.AVRO.value,
                ContentEncoding.IDENTITY.value,
                {},
            ),
            (
                f"{self.temp_dir}/test.orc",
                ContentType.ORC.value,
                ContentEncoding.IDENTITY.value,
                {},
            ),
        ]

        filesystems = [
            ("fsspec", self.fsspec_fs),
            ("pyarrow", self.pyarrow_fs),
            ("auto-inferred", None),
        ]

        for path, content_type, content_encoding, extra_kwargs in test_cases:
            for fs_name, filesystem in filesystems:
                with self.subTest(
                    content_type=content_type,
                    filesystem=fs_name,
                    encoding=content_encoding,
                ):
                    result = file_to_dataframe(
                        path=path,
                        content_type=content_type,
                        content_encoding=content_encoding,
                        filesystem=filesystem,
                        **extra_kwargs,
                    )
                    self._assert_dataframes_equal(result, self.test_data)

    def test_compression_encoding_with_different_filesystems(self):
        """Test that compression encoding works correctly with different filesystem types."""
        test_cases = [
            (f"{self.temp_dir}/test.csv", ContentEncoding.IDENTITY.value),
            (f"{self.temp_dir}/test_gzip.csv.gz", ContentEncoding.GZIP.value),
            (f"{self.temp_dir}/test_bzip2.csv.bz2", ContentEncoding.BZIP2.value),
        ]

        filesystems = [
            ("fsspec", self.fsspec_fs),
            ("pyarrow", self.pyarrow_fs),
            ("auto-inferred", None),
        ]

        for path, content_encoding in test_cases:
            for fs_name, filesystem in filesystems:
                with self.subTest(encoding=content_encoding, filesystem=fs_name):
                    result = file_to_dataframe(
                        path=path,
                        content_type=ContentType.CSV.value,
                        content_encoding=content_encoding,
                        filesystem=filesystem,
                        has_header=True,
                    )
                    self._assert_dataframes_equal(result, self.test_data)

    def test_filesystem_open_kwargs(self):
        """Test that filesystem open kwargs are properly passed through."""
        from deltacat.utils.polars import read_csv

        # Test with custom fs_open_kwargs
        result = read_csv(
            f"{self.temp_dir}/test.csv",
            filesystem=self.fsspec_fs,
            content_encoding=ContentEncoding.IDENTITY.value,
            fs_open_kwargs={
                "encoding": "utf-8"
            },  # This should be passed to filesystem.open()
            has_header=True,
        )
        self._assert_dataframes_equal(result, self.test_data)



def ExperimentalAPI(obj):
    """Decorator for documenting experimental APIs.

    Experimental APIs are classes and methods that are in development and may
    change at any time in their development process. You should not expect
    these APIs to be stable until their tag is changed to `DeveloperAPI` or
    `PublicAPI`.

    Subclasses that inherit from a ``@ExperimentalAPI`` base class can be
    assumed experimental as well.

    This decorator has no effect on runtime behavior
    """
    return obj


def DeveloperAPI(obj):
    """Decorator for documenting experimental APIs.

    Developer APIs are classes and methods explicitly exposed to developers
    for low level integrations with DeltaCAT (e.g.: compute engines, other catalogs).
    You can generally expect these APIs to be stable sans minor changes (but less stable than public APIs).

    This decorator has no effect on runtime behavior
    """
    return obj


def PublicAPI(obj):
    """Decorator for documenting public APIs.

    Public APIs are classes and methods exposed to end users which are expected to remain stable across releases.

    This decorator has no effect on runtime behavior
    """
    return obj



import io

import pytest
from faker import Faker

from deltacat.experimental.storage.rivulet.schema.datatype import Datatype
from deltacat.experimental.storage.rivulet.mvp.Table import MvpTable
from deltacat.experimental.storage.rivulet.schema.schema import Schema
import random
import string
from PIL import Image

FIXTURE_ROW_COUNT = 10000


@pytest.fixture
def ds1_dataset() -> MvpTable:
    """
    dataset with one million rows
    primary key is integer between 1 and 1,000,000

    TODO change to user Faker instead of int ranges
    """

    # Function to generate random names
    def generate_random_name():
        return "".join(
            random.choices(
                string.ascii_uppercase + string.ascii_lowercase, k=random.randint(3, 10)
            )
        )

    # Create a list of numbers from 1 to TEST_ROW_COUNT
    ids = list(range(1, FIXTURE_ROW_COUNT + 1))
    random.shuffle(ids)

    # Generate one million rows
    return MvpTable(
        {
            "id": ids,
            "name": [generate_random_name() for _ in range(FIXTURE_ROW_COUNT)],
            "age": [random.randint(18, 100) for _ in range(FIXTURE_ROW_COUNT)],
        }
    )


@pytest.fixture
def ds1_schema():
    return Schema(
        {
            ("id", Datatype.int32()),
            ("name", Datatype.string()),
            ("age", Datatype.int32()),
        },
        "id",
    )


@pytest.fixture
def ds2_dataset():
    """
    dataset2 with one million rows that can be joined to ds1
    primary key is integer between 1 and 1,000,000
    """
    # Create a list of numbers from 1 to 1,000,000
    ids = list(range(1, FIXTURE_ROW_COUNT + 1))
    random.shuffle(ids)

    fake = Faker()

    # Generate one million rows
    return MvpTable(
        {
            "id": ids,
            "address": [fake.address() for _ in range(FIXTURE_ROW_COUNT)],
            "zip": [fake.zipcode() for _ in range(FIXTURE_ROW_COUNT)],
        }
    )


@pytest.fixture
def ds2_schema():
    return Schema(
        {
            ("id", Datatype.int32()),
            ("address", Datatype.string()),
            ("zip", Datatype.string()),
        },
        "id",
    )


@pytest.fixture
def combined_schema(ds1_schema, ds2_schema):
    return Schema(
        {
            ("id", Datatype.int32()),
            ("address", Datatype.string()),
            ("zip", Datatype.string()),
            ("name", Datatype.string()),
            ("age", Datatype.int32()),
        },
        "id",
    )


@pytest.fixture
def dataset_images_with_label() -> (MvpTable, Schema):
    """
    dataset with one thousand images and labels, generated dynamically
    primary key is integer between 1 and 1,000
    """
    ROW_COUNT = 1000
    fake = Faker()
    schema = Schema(
        {
            ("id", Datatype.int32()),
            ("image", Datatype.image("jpg")),
            ("label", Datatype.string()),
        },
        "id",
    )

    # Create a list of numbers from 1 to ROW_COUNT
    ids = list(range(1, ROW_COUNT + 1))
    random.shuffle(ids)

    fake_image = Image.new(
        "RGB",
        (512, 512),
        color=(random.randint(0, 255), random.randint(0, 255), random.randint(0, 255)),
    )
    # get bytes from image encoded as png
    buffer = io.BytesIO()
    fake_image.save(buffer, format="PNG")
    # seek to start of buffer since we just wrote to it
    buffer.seek(0)
    image_bytes = buffer.read()
    # Generate one million rows
    return (
        MvpTable(
            {
                "id": ids,
                "image": [image_bytes for _ in range(ROW_COUNT)],
                "label": [fake.name() for _ in range(ROW_COUNT)],
            }
        ),
        schema,
    )



import pytest

from deltacat.storage.model.shard import ShardingStrategy


def test_sharding_strategy_from_string_range():
    """
    Tests that from_string('range') returns an instance of RangeShardingStrategy.
    """
    from deltacat.experimental.storage.rivulet.shard.range_shard import (
        RangeShardingStrategy,
    )

    strategy = ShardingStrategy.from_string("range")
    assert isinstance(strategy, RangeShardingStrategy)


def test_sharding_strategy_from_string_invalid():
    """
    Tests that from_string(...) raises ValueError for an unknown strategy string.
    """
    with pytest.raises(ValueError) as exc_info:
        ShardingStrategy.from_string("unknown_strategy")
    assert "Unsupported sharding strategy type: unknown_strategy" in str(exc_info.value)



import argparse
import pathlib

from deltacat.compute import (
    job_client,
    JobStatus,
)


def run_async(
    source: str,
    dest: str,
    jobs_to_submit: int,
    job_timeout: int,
    cloud: str,
    restart_ray: bool,
):
    # print package version info
    working_dir = pathlib.Path(__file__).parent
    cluster_cfg_file_path = working_dir.joinpath(cloud).joinpath("deltacat.yaml")
    job_number = 0
    client = job_client(cluster_cfg_file_path, restart_ray=restart_ray)
    job_ids = []
    while jobs_to_submit > 0:
        jobs_to_submit -= 1
        job_dest = dest + f".{job_number}"
        job_id = client.submit_job(
            # Entrypoint shell command to execute
            entrypoint=f"python3 indexer.py --source '{source}' --dest '{job_dest}'",
            # Path to the local directory that contains the indexer.py file
            # This entire directory will be zipped into a job package, so keep
            # it small.
            runtime_env={"working_dir": working_dir},
        )
        job_ids.append(job_id)
        job_number += 1

    print("Waiting for all jobs to complete...")
    job_number = 0
    all_job_logs = ""
    for job_id in job_ids:
        job_status = client.await_job(job_id, timeout_seconds=job_timeout)
        if job_status != JobStatus.SUCCEEDED:
            print(f"Job `{job_id}` logs: ")
            print(client.get_job_logs(job_id))
            raise RuntimeError(f"Job `{job_id}` terminated with status: {job_status}")
        all_job_logs += f"\nJob #{job_number} logs: \n"
        all_job_logs += client.get_job_logs(job_id)
        job_number += 1
    print("All jobs completed!")
    print("Job Logs: ")
    print(all_job_logs)


def run_sync(
    source: str,
    dest: str,
    jobs_to_submit: int,
    job_timeout: int,
    cloud: str,
    restart_ray: bool,
):
    working_dir = pathlib.Path(__file__).parent
    cluster_cfg_file_path = working_dir.joinpath(cloud).joinpath("deltacat.yaml")
    client = job_client(cluster_cfg_file_path, restart_ray=restart_ray)
    job_number = 0
    while job_number < jobs_to_submit:
        job_dest = dest + f".{job_number}"
        job_run_result = client.run_job(
            # Entrypoint shell command to execute
            entrypoint=f"python3 indexer.py --source '{source}' --dest '{job_dest}'",
            # Path to the local directory that contains the indexer.py file
            # This entire directory will be zipped into a job package, so keep
            # it small.
            runtime_env={"working_dir": working_dir},
            timeout_seconds=job_timeout,
        )
        print(
            f"Job ID {job_run_result.job_id} terminal state: {job_run_result.job_status}"
        )
        print(f"Job ID {job_run_result.job_id} logs: ")
        print(job_run_result.job_logs)
        job_number += 1


def run(
    source: str,
    dest: str,
    restart_ray: bool,
    jobs_to_submit: int,
    job_timeout: int,
    asynchronous: bool,
    cloud_provider: str,
):
    run_func = run_async if asynchronous else run_sync
    run_func(
        source=source,
        dest=dest,
        jobs_to_submit=jobs_to_submit,
        job_timeout=job_timeout,
        cloud=cloud_provider,
        restart_ray=restart_ray,
    )


if __name__ == "__main__":
    """
    # Run this example through a command of the form:
    $ python ./deltacat/examples/job_runner.py -- \
    $ --source text+s3://openalex-mag-format/data_dump_v1/2022-07-08/nlp/PaperAbstractsInvertedIndex.txt_part31 \
    $ --dest parquet+s3://deltacat-example-output/openalex/PaperAbstractsInvertedIndex.part31.parquet \
    $ --asynchronous \
    $ --jobs-to-submit 100 \
    $ --job-timeout 90 \
    $ --cloud-provider aws
    """
    script_args = [
        (
            [
                "--source",
            ],
            {
                "help": "Source DeltaCAT URL to index.",
                "type": str,
                "default": "text+s3://openalex-mag-format/data_dump_v1/2022-07-08/nlp/PaperAbstractsInvertedIndex.txt_part31",
            },
        ),
        (
            [
                "--dest",
            ],
            {
                "help": "Destination DeltaCAT URL to store the indexed file.",
                "type": str,
                "default": "parquet+s3://deltacat-example-output/openalex/PaperAbstractsInvertedIndex.part31.parquet",
            },
        ),
        (
            [
                "--restart-ray",
            ],
            {
                "help": "Restart Ray on an existing cluster.",
                "action": "store_true",
                "default": False,
            },
        ),
        (
            [
                "--asynchronous",
            ],
            {
                "help": "Run jobs asynchronously.",
                "action": "store_true",
                "default": False,
            },
        ),
        (
            [
                "--jobs-to-submit",
            ],
            {
                "help": "Number of indexer jobs to submit for execution.",
                "type": int,
                "default": 1,
            },
        ),
        (
            [
                "--job-timeout",
            ],
            {
                "help": "Job timeout in seconds.",
                "type": int,
                "default": 300,
            },
        ),
        (
            [
                "--cloud-provider",
            ],
            {
                "help": "Ray Cluster Cloud Provider ('aws' or 'gcp')",
                "type": str,
                "default": "aws",
            },
        ),
    ]

    # parse CLI input arguments
    parser = argparse.ArgumentParser()
    for args, kwargs in script_args:
        parser.add_argument(*args, **kwargs)
    args = parser.parse_args()
    print(f"Command Line Arguments: {args}")

    # run the example using os.environ as kwargs
    run(**vars(args))



from __future__ import annotations
from typing import Dict, List, Any, Optional, Tuple
from pyiceberg.manifest import DataFile

# Type aliases to simplify nested types
DataFileWithSequence = Tuple[int, DataFile]  # (sequence_number, data_file)
DataFileList = List[DataFileWithSequence]  # List of data files with sequence numbers
DataFileListGroup = List[DataFileList]  # Group of data file lists


class ConvertInputFiles(Dict):
    @staticmethod
    def of(
        partition_value: Any,
        all_data_files_for_dedupe: Optional[DataFileList] = None,
        applicable_data_files: Optional[DataFileListGroup] = None,
        applicable_equality_delete_files: Optional[DataFileListGroup] = None,
        existing_position_delete_files: Optional[DataFileList] = None,
    ) -> ConvertInputFiles:

        result = ConvertInputFiles()
        result["partition_value"] = partition_value
        result["all_data_files_for_dedupe"] = all_data_files_for_dedupe
        result["applicable_data_files"] = applicable_data_files
        result["applicable_equality_delete_files"] = applicable_equality_delete_files
        result["existing_position_delete_files"] = existing_position_delete_files
        return result

    @property
    def partition_value(self) -> Any:
        return self["partition_value"]

    @property
    def all_data_files_for_dedupe(self) -> Optional[DataFileList]:
        return self["all_data_files_for_dedupe"]

    @property
    def applicable_data_files(self) -> Optional[DataFileListGroup]:
        return self["applicable_data_files"]

    @property
    def applicable_equality_delete_files(
        self,
    ) -> Optional[DataFileListGroup]:
        return self["applicable_equality_delete_files"]

    @property
    def existing_position_delete_files(self) -> Optional[DataFileList]:
        return self["existing_position_delete_files"]

    @partition_value.setter
    def partition_value(self, partition_value: Any) -> None:
        self["partition_value"] = partition_value

    @all_data_files_for_dedupe.setter
    def all_data_files_for_dedupe(
        self, all_data_files_for_dedupe: Optional[DataFileList]
    ) -> None:
        self["all_data_files_for_dedupe"] = all_data_files_for_dedupe

    @applicable_data_files.setter
    def applicable_data_files(
        self, applicable_data_files: Optional[DataFileListGroup]
    ) -> None:
        self["applicable_data_files"] = applicable_data_files

    @applicable_equality_delete_files.setter
    def applicable_equality_delete_files(
        self,
        applicable_equality_delete_files: Optional[DataFileListGroup],
    ) -> None:
        self["applicable_equality_delete_files"] = applicable_equality_delete_files

    @existing_position_delete_files.setter
    def existing_position_delete_files(
        self, existing_position_delete_files: Optional[DataFileList]
    ) -> None:
        self["existing_position_delete_files"] = existing_position_delete_files



import posixpath
from deltacat.utils.metafile_locator import _find_partition_path
import pytest

import pyarrow as pa
from deltacat.experimental.storage.rivulet import Schema, Field, Datatype
from deltacat.experimental.storage.rivulet.dataset import Dataset
from deltacat.experimental.storage.rivulet.reader.query_expression import (
    QueryExpression,
)


@pytest.fixture
def sample_schema():
    return Schema(
        fields=[
            Field("id", Datatype.int32(), is_merge_key=True),
            Field("name", Datatype.string()),
            Field("age", Datatype.int32()),
        ]
    )


@pytest.fixture
def sample_pydict():
    return {"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"], "age": [25, 30, 35]}


@pytest.fixture
def sample_parquet_data(tmp_path, sample_pydict):
    parquet_path = tmp_path / "test.parquet"
    table = pa.Table.from_pydict(sample_pydict)
    pa.parquet.write_table(table, parquet_path)
    return parquet_path


# Updated Tests


def test_dataset_creation_with_schema(tmp_path, sample_schema):
    dataset = Dataset(dataset_name="test_dataset", schema=sample_schema)
    assert len(dataset.fields) == 3
    assert "id" in dataset.fields
    assert dataset.fields["id"].is_merge_key


def test_dataset_initialization_with_metadata(tmp_path):
    dataset = Dataset(dataset_name="test_dataset", metadata_uri=str(tmp_path))
    assert dataset.dataset_name == "test_dataset"
    assert dataset._metadata_folder.startswith(".riv-meta")


def test_invalid_dataset_initialization():
    with pytest.raises(ValueError, match="Name must be a non-empty string"):
        Dataset(dataset_name="")


def test_dataset_creation_metadata_structure(tmp_path):
    dataset = Dataset(dataset_name="test_dataset", metadata_uri=str(tmp_path))

    assert dataset._metadata_folder.startswith(".riv-meta")
    assert dataset._namespace == "default"
    assert dataset.dataset_name == "test_dataset"
    assert dataset._metadata_path == str(tmp_path / ".riv-meta-test_dataset")

    locator = dataset._locator
    root_uri = dataset._metadata_path

    partition_path = _find_partition_path(root_uri, locator)

    # Ensures that directory structure for namespace -> table -> table_version -> stream_id -> partition_id exists
    assert posixpath.exists(partition_path)


def test_fields_accessor_add_field(tmp_path, sample_schema):
    dataset = Dataset(dataset_name="test_dataset", schema=sample_schema)
    dataset.fields.add("new_field", Datatype.float())
    assert "new_field" in dataset.fields
    assert dataset.fields["new_field"].datatype == Datatype.float()

    dataset.fields["new_field2"] = Field("new_field2", Datatype.int32())
    assert "new_field2" in dataset.fields
    assert "new_field2" in dataset.schemas["all"]
    with pytest.raises(TypeError):
        dataset.fields["new_field3"] = 2


def test_field_removal(tmp_path, sample_schema):
    dataset = Dataset(dataset_name="test_dataset", schema=sample_schema)
    del dataset.fields["age"]
    assert "age" not in dataset.fields
    with pytest.raises(ValueError):
        del dataset.fields["age"]
    with pytest.raises(KeyError):
        _ = dataset.fields["age"]


def test_fields_accessor_repr(tmp_path, sample_schema):
    dataset = Dataset(dataset_name="test_dataset", schema=sample_schema)
    repr_output = repr(dataset.fields)
    for field_name in ["id", "name", "age"]:
        assert field_name in repr_output, f"Field '{field_name}' missing in repr output"


def test_schemas_accessor_add_group(tmp_path, sample_schema):
    dataset = Dataset(dataset_name="test_dataset", schema=sample_schema)
    dataset.schemas["analytics"] = ["id", "name"]
    assert "analytics" in dataset.schemas
    assert len(dataset.schemas["analytics"]) == 2


def test_schema_removal(tmp_path, sample_schema):
    dataset = Dataset(dataset_name="test_dataset", schema=sample_schema)
    with pytest.raises(ValueError):
        del dataset.schemas["all"]
    with pytest.raises(ValueError):
        del dataset.schemas["does_not_exist"]
    dataset.schemas["new"] = ["id", "name"]
    del dataset.schemas["new"]
    with pytest.raises(KeyError):
        _ = dataset.schemas["new"]


def test_dataset_from_parquet(tmp_path, sample_parquet_data):
    dataset = Dataset.from_parquet(
        name="test_dataset",
        file_uri=str(sample_parquet_data),
        metadata_uri=str(tmp_path),
        merge_keys="id",
    )
    assert len(dataset.fields) == 3
    assert "id" in dataset.fields
    assert dataset.fields["id"].is_merge_key


def test_parquet_schema_modes(tmp_path, sample_pydict):
    # Create two parquet files with overlapping and unique schemas
    data_1 = {"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]}
    data_2 = {"id": [4, 5, 6], "age": [25, 30, 35]}

    path_1 = tmp_path / "data1.parquet"
    path_2 = tmp_path / "data2.parquet"
    pa.parquet.write_table(pa.Table.from_pydict(data_1), path_1)
    pa.parquet.write_table(pa.Table.from_pydict(data_2), path_2)

    dataset_union = Dataset.from_parquet(
        name="test_dataset_union",
        file_uri=str(tmp_path),
        merge_keys="id",
        schema_mode="union",
    )
    assert len(dataset_union.fields) == 3  # id, name, age

    dataset_intersect = Dataset.from_parquet(
        name="test_dataset_intersect",
        file_uri=str(tmp_path),
        merge_keys="id",
        schema_mode="intersect",
    )
    assert len(dataset_intersect.fields) == 1  # Only id


def test_merge_all_schemas():
    schema1 = Schema(
        fields=[
            Field("id", Datatype.int32(), is_merge_key=True),
            Field("name", Datatype.string()),
        ]
    )
    schema2 = Schema(
        fields=[
            Field("id", Datatype.int32(), is_merge_key=True),
            Field("age", Datatype.int32()),
        ]
    )
    merged_schema = Schema.merge_all([schema1, schema2])
    assert len(merged_schema) == 3
    assert "id" in merged_schema
    assert "name" in merged_schema
    assert "age" in merged_schema


def test_writer_creation_with_custom_format(tmp_path, sample_schema):
    dataset = Dataset(dataset_name="test_dataset", schema=sample_schema)
    writer = dataset.writer(file_format="feather")
    assert writer is not None


def test_scan_with_query(tmp_path, sample_schema):
    dataset = Dataset(dataset_name="test_dataset", schema=sample_schema)
    query = QueryExpression()  # Placeholder query
    scan = dataset.scan(query)
    assert scan is not None


def test_add_schema_to_new_schemas(tmp_path):
    """Test adding a schema to a new field group."""
    base_uri = str(tmp_path / "test_dataset")
    dataset = Dataset(dataset_name=base_uri)

    schema = Schema(
        [
            ("id", Datatype.int32()),
            ("name", Datatype.string()),
            ("age", Datatype.int32()),
        ],
        merge_keys=["id"],
    )

    dataset.add_schema(schema, schema_name="new_group")

    # Verify the field group is added
    assert "new_group" in dataset.schemas
    assert len(dataset.schemas["new_group"]) == 3
    assert dataset.schemas["new_group"]["id"].datatype == Datatype.int32()
    assert dataset.schemas["new_group"]["name"].datatype == Datatype.string()
    assert dataset.schemas["new_group"]["age"].datatype == Datatype.int32()


def test_add_schema_to_existing_schemas(tmp_path):
    """Test merging a schema into an existing field group."""
    base_uri = str(tmp_path / "test_dataset")
    dataset = Dataset(dataset_name=base_uri)

    schema_1 = Schema(
        [
            ("id", Datatype.int32()),
            ("name", Datatype.string()),
        ],
        merge_keys=["id"],
    )

    dataset.add_schema(schema_1, schema_name="existing_group")

    schema_2 = Schema(
        [
            ("age", Datatype.int32()),
            ("email", Datatype.string()),
        ],
        merge_keys=["id"],
    )

    dataset.add_schema(schema_2, schema_name="existing_group")

    # Verify the merged schema
    assert "existing_group" in dataset.schemas
    assert len(dataset.schemas["existing_group"]) == 4
    assert dataset.schemas["existing_group"]["id"].datatype == Datatype.int32()
    assert dataset.schemas["existing_group"]["name"].datatype == Datatype.string()
    assert dataset.schemas["existing_group"]["age"].datatype == Datatype.int32()
    assert dataset.schemas["existing_group"]["email"].datatype == Datatype.string()


def test_add_schema_conflicting_fields(tmp_path):
    """Test adding a schema with conflicting fields."""
    base_uri = str(tmp_path / "test_dataset")
    dataset = Dataset(dataset_name=base_uri)

    schema_1 = Schema(
        [
            ("id", Datatype.int32()),
            ("name", Datatype.string()),
        ],
        merge_keys=["id"],
    )

    dataset.add_schema(schema_1, schema_name="conflicting_group")

    schema_2 = Schema(
        [
            ("id", Datatype.string()),  # Conflict: datatype mismatch
            ("age", Datatype.int32()),
        ],
        merge_keys=["id"],
    )

    with pytest.raises(ValueError, match="already exists"):
        dataset.add_schema(schema_2, schema_name="conflicting_group")

    schema_3 = Schema(
        [
            ("id", Datatype.int32()),  # Conflict: datatype mismatch
            ("age", Datatype.int32()),
        ],
        merge_keys=["id"],
    )

    dataset.add_schema(schema_3, schema_name="conflicting_group")
    assert "conflicting_group" in dataset.schemas
    assert len(dataset.schemas["conflicting_group"]) == 3
    assert dataset.schemas["conflicting_group"]["id"].datatype == Datatype.int32()
    assert dataset.schemas["conflicting_group"]["name"].datatype == Datatype.string()
    assert dataset.schemas["conflicting_group"]["age"].datatype == Datatype.int32()


def test_add_fields_with_merge_key_field(tmp_path):
    base_uri = str(tmp_path / "test_dataset")
    dataset = Dataset(dataset_name=base_uri)
    dataset.add_fields([Field("my_merge_key", Datatype.string(), True)])
    assert dataset.schemas["default"].get_merge_key() == "my_merge_key"


def test_add_schema_to_nonexistent_schemas(tmp_path):
    """Test adding a schema to a nonexistent field group."""
    base_uri = str(tmp_path / "test_dataset")
    dataset = Dataset(dataset_name=base_uri)

    schema = Schema(
        [
            ("id", Datatype.int32()),
            ("name", Datatype.string()),
        ],
        merge_keys=["id"],
    )

    # Add to a non-existent field group
    dataset.add_schema(schema, schema_name="nonexistent_group")

    # Verify the field group is created
    assert "nonexistent_group" in dataset.schemas
    assert len(dataset.schemas["nonexistent_group"]) == 2


def test_add_missing_field_to_schema_raises_error(tmp_path, sample_schema):
    """
    Test that attempting to add a missing field to the 'all' schema raises a ValueError.
    """
    dataset = Dataset(dataset_name="test_dataset", schema=sample_schema)

    # Attempt to add a non-existent field to the 'all' schema
    with pytest.raises(
        ValueError, match="Field 'missing_field' does not exist in the dataset."
    ):
        dataset.schemas["all"] = [
            "missing_field"
        ]  # Attempt to set a list with a missing field


def test_schemas_accessor_methods(tmp_path, sample_schema):
    """
    Test the __iter__, __len__, and __repr__ methods of SchemasAccessor.
    """
    dataset = Dataset(
        dataset_name="test_dataset", schema=sample_schema
    )  # Default schema is defined automatically
    dataset.schemas["schema_1"] = ["id", "name"]
    dataset.schemas["schema_2"] = ["age"]

    # Test __iter__
    schema_names = list(iter(dataset.schemas))
    assert set(schema_names) == {
        "schema_1",
        "schema_2",
        "all",
        "default",
    }, "Schema names do not match expected values"

    # Test __len__
    assert len(dataset.schemas) == 4, "Length of schemas accessor is incorrect"

    # Test __repr__
    repr_output = repr(dataset.schemas)
    for schema_name in ["schema_1", "schema_2", "all"]:
        assert (
            schema_name in repr_output
        ), f"Schema '{schema_name}' missing in repr output"


def test_get_merge_keys(tmp_path, sample_schema):
    """
    Test the get_merge_keys method to ensure it returns all merge keys in the dataset.
    """
    dataset = Dataset(dataset_name="test_dataset", schema=sample_schema)

    # Add fields with additional merge key to the dataset
    other_schema = Schema(
        [("id2", Datatype.int32()), ("zip", Datatype.string())], merge_keys=["id2"]
    )

    dataset.add_schema(other_schema, "id2+zip")

    # Call get_merge_keys and validate the result
    merge_keys = dataset.get_merge_keys()
    assert merge_keys == [
        "id",
        "id2",
    ], f"Expected merge keys ['id', 'id2'], got {merge_keys}"


def test_add_fields_no_fields_raises_error(tmp_path, sample_schema):
    dataset = Dataset(dataset_name="test_dataset")
    with pytest.raises(ValueError):
        dataset.add_fields(fields=[])


def test_add_fields_mismatched_merge_keys_raises_error(tmp_path, sample_schema):
    dataset = Dataset(dataset_name="test_dataset")
    with pytest.raises(
        ValueError,
        match="The following merge keys were not found in the provided fields: does_not_exist",
    ):
        dataset.add_fields(fields=sample_schema.values(), merge_keys=["does_not_exist"])

    with pytest.raises(TypeError, match="Merge key status conflict"):
        dataset.add_fields(
            fields=[Field("id", Datatype.int32()), Field("name", Datatype.string())],
            merge_keys=["id"],
        )
